#!/usr/bin/env python

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2022 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from collections import namedtuple

from pya2l.utils import SingletonBase


def camel_case(name, upper_first=True):
    splitty = [n.lower() for n in name.split("_")]
    result = []
    result.append(splitty[0])
    if len(splitty) > 1:
        for part in splitty[1:]:
            xxx = f"{part[0].upper()}{part[1:]}"
            result.append(xxx)
    result = "".join(result)
    if upper_first:
        result = f"{result[0].upper()}{result[1:]}"
    return result


class MULTIPLE(SingletonBase):
    pass


class Uint(SingletonBase):
    pass


class Int(SingletonBase):
    pass


class Ulong(SingletonBase):
    pass


class Long(SingletonBase):
    pass


class Float(SingletonBase):
    pass


class String(SingletonBase):
    pass


class Enum(SingletonBase):
    pass


class Ident(SingletonBase):
    pass


class Datatype(SingletonBase):
    enumValues = (
        "UBYTE",
        "SBYTE",
        "UWORD",
        "SWORD",
        "ULONG",
        "SLONG",
        "A_UINT64",
        "A_INT64",
        "FLOAT16_IEEE",
        "FLOAT32_IEEE",
        "FLOAT64_IEEE",
    )


class Datasize(SingletonBase):
    enumValues = ("BYTE", "WORD", "LONG")


class Addrtype(SingletonBase):
    enumValues = ("PBYTE", "PWORD", "PLONG", "DIRECT")


class Byteorder(SingletonBase):
    enumValues = ("LITTLE_ENDIAN", "BIG_ENDIAN", "MSB_LAST", "MSB_FIRST")


class Indexorder(SingletonBase):
    enumValues = ("INDEX_INCR", "INDEX_DECR")


class Linktype(SingletonBase):
    enumValues = ("SYMBOL_TYPE_LINK",)


CompuPair = namedtuple("CompuPair", "inVal outVal")
CompuTriplet = namedtuple("CompuTriplet", "valMin valMax outVal")


class KeywordType(type):
    classDict = {}
    classes = set()

    def __new__(metacls, name, bases, namespace):
        newKlass = super(metacls, KeywordType).__new__(metacls, name, bases, namespace)
        KeywordType.classDict[newKlass.__name__] = newKlass
        KeywordType.classes.add(newKlass.__name__.lower())

        attrs = namespace.get("attrs", [])

        fixedAttributes = []
        variableAttribute = None
        if attrs:
            fixedAttributes = [attr[1] for attr in attrs if MULTIPLE not in attr]
            variableAttribute = [attr[1] for attr in attrs if MULTIPLE in attr]
            if variableAttribute:
                # print("VA: {} ==> {}".format(newKlass.__name__, variableAttribute))
                variableAttribute = variableAttribute[0]
        newKlass.fixedAttributes = fixedAttributes
        newKlass.variableAttribute = variableAttribute
        newKlass.attrDict = dict(zip([a[1] for a in attrs], attrs))

        return newKlass

    def __getitem__(cls, key):
        return cls.attrDict.get(key, None)

    def __eq__(cls, other):
        return cls.__class__.__name__ == other.__class__.__name__

    def __lt__(cls, other):
        return cls.__class__.__name__ < other.__class__.__name__

    def __hash__(cls):
        return id(cls)

    @classmethod
    def getClass(metacls, name):
        result = KeywordType.classDict.get(name, None)
        if not result:
            pass
        return result


class Keyword(metaclass=KeywordType):

    multiple = False
    block = False
    optional = True
    textNode = False
    attrs = []
    children = []

    @classmethod
    def attributeNames(cls):
        return [attr[1] for attr in cls.attrs if MULTIPLE not in attr]

    def __str__(self):
        return f"< {self.__class__.__name__} @{id(self):0X} >"

    @classmethod
    def class_name(cls):
        return cls.__name__

    @classmethod
    def lower_name(cls):
        return cls.class_name().lower()

    @classmethod
    def plural_name(cls):
        name = cls.lower_name()
        return "{}{}".format(name, "s" if name[-1] != "s" else "")

    @classmethod
    def assoc_name(cls):
        return cls.plural_name() if cls.multiple else cls.lower_name()

    @classmethod
    def camel_case_name(cls, upper_first=True):
        return camel_case(cls.__name__, upper_first)

    @classmethod
    def camel_case_plural_name(cls, upper_first=True):
        name = cls.camel_case_name(upper_first)
        return "{}{}".format(name, "s" if name[-1] != "s" else "")


##
##
##
class A2ML(Keyword):
    # multiple = True
    block = True
    # Contains AML code of interface specific description data.


class A2ML_VERSION(Keyword):
    attrs = [
        (Uint, "VersionNo"),  # Version number of AML part.
        (Uint, "UpgradeNo"),  # Upgrade number of AML part.
    ]


class ADDR_EPK(Keyword):
    multiple = True
    attrs = [(Ulong, "Address")]  # Address of the EPROM identifier.


class ALIGNMENT_BYTE(Keyword):
    attrs = [
        (
            Uint,
            "AlignmentBorder",
        )  # Describes the border at which the value is aligned to,
        # i.e. its memory address must be dividable by the value AlignmentBorder.)
    ]


class ALIGNMENT_FLOAT16_IEEE(Keyword):
    attrs = [
        (
            Uint,
            "AlignmentBorder",
        )  # Describes the border at which the value is aligned to,
        # i.e. its memory address must be dividable by the value AlignmentBorder.)
    ]


class ALIGNMENT_FLOAT32_IEEE(Keyword):
    attrs = [
        (
            Uint,
            "AlignmentBorder",
        )  # Describes the border at which the value is aligned to,
        # i.e. its memory address must be dividable by the value AlignmentBorder.)
    ]


class ALIGNMENT_FLOAT64_IEEE(Keyword):
    attrs = [
        (
            Uint,
            "AlignmentBorder",
        )  # Describes the border at which the value is aligned to,
        # i.e. its memory address must be dividable by the value AlignmentBorder.)
    ]


class ALIGNMENT_INT64(Keyword):
    attrs = [
        (
            Uint,
            "AlignmentBorder",
        )  # Describes the border at which the value is aligned to,
        # i.e. its memory address must be dividable by the value AlignmentBorder.)
    ]


class ALIGNMENT_LONG(Keyword):
    attrs = [
        (
            Uint,
            "AlignmentBorder",
        )  # Describes the border at which the value is aligned to,
        # i.e. its memory address must be dividable by the value AlignmentBorder.)
    ]


class ALIGNMENT_WORD(Keyword):
    attrs = [
        (
            Uint,
            "AlignmentBorder",
        )  # Describes the border at which the value is aligned to,
        # i.e. its memory address must be dividable by the value AlignmentBorder.)
    ]


class ANNOTATION(Keyword):
    multiple = True
    block = True
    children = ["ANNOTATION_LABEL", "ANNOTATION_ORIGIN", "ANNOTATION_TEXT"]


class ANNOTATION_LABEL(Keyword):
    attrs = [(String, "Label")]  # label or title of the annotation


class ANNOTATION_ORIGIN(Keyword):
    attrs = [(String, "Origin")]  # creator or creating system of the annotation


class ANNOTATION_TEXT(Keyword):
    block = True
    textNode = True

    attrs = [(String, "Text", MULTIPLE)]


class ARRAY_SIZE(Keyword):
    attrs = [
        (
            Uint,
            "Number",
        )  # Number of measurement values included in respective measurement
        # object  (maximum    value  of  ‘Number’: 32767).
    ]
    # The use of this keyword should be replaced by MATRIX_DIM.


class ASAP2_VERSION(Keyword):
    attrs = [
        (Uint, "VersionNo"),  # Version number of ASAM MCD-2MC standard.
        (Uint, "UpgradeNo"),  # Upgrade number of ASAM MCD-2MC standard.
    ]


class AXIS_DESCR(Keyword):
    multiple = True
    block = True
    children = [
        "ANNOTATION",
        "AXIS_PTS_REF",
        "BYTE_ORDER",
        "CURVE_AXIS_REF",
        "DEPOSIT",
        "EXTENDED_LIMITS",
        "FIX_AXIS_PAR",
        "FIX_AXIS_PAR_DIST",
        "FIX_AXIS_PAR_LIST",
        "FORMAT",
        "MAX_GRAD",
        "MONOTONY",
        "PHYS_UNIT",
        "READ_ONLY",
        "STEP_SIZE",
    ]
    attrs = [
        (
            Enum,
            "Attribute",
            ("CURVE_AXIS", "COM_AXIS", "FIX_AXIS", "RES_AXIS", "STD_AXIS"),
        ),
        (Ident, "InputQuantity"),
        (Ident, "Conversion"),
        (Uint, "MaxAxisPoints"),
        (Float, "LowerLimit"),
        (Float, "UpperLimit"),
    ]


class AXIS_PTS(Keyword):
    multiple = True
    block = True
    children = [
        "ANNOTATION",
        "BYTE_ORDER",
        "CALIBRATION_ACCESS",
        "DEPOSIT",
        "DISPLAY_IDENTIFIER",
        "ECU_ADDRESS_EXTENSION",
        "EXTENDED_LIMITS",
        "FORMAT",
        "FUNCTION_LIST",
        "GUARD_RAILS",
        "IF_DATA",
        "MONOTONY",
        "PHYS_UNIT",
        "READ_ONLY",
        "REF_MEMORY_SEGMENT",
        "STEP_SIZE",
        "SYMBOL_LINK",
    ]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Ulong, "Address"),
        (Ident, "InputQuantity"),
        (Ident, "DepositAttr"),
        (Float, "MaxDiff"),
        (Ident, "Conversion"),
        (Uint, "MaxAxisPoints"),
        (Float, "LowerLimit"),
        (Float, "UpperLimit"),
    ]


class AXIS_PTS_REF(Keyword):
    attrs = [
        (
            Ident,
            "AxisPoints",
        )  # Name of the AXIS_PTS data record which describes the axis points
        # distribution (group axis points and record layout: see AXIS_PTS).
    ]


class AXIS_PTS_X(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class AXIS_PTS_Y(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class AXIS_PTS_Z(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class AXIS_PTS_4(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class AXIS_PTS_5(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class AXIS_RESCALE_X(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Uint, "MaxNumberOfRescalePairs"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class AXIS_RESCALE_Y(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Uint, "MaxNumberOfRescalePairs"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class AXIS_RESCALE_Z(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Uint, "MaxNumberOfRescalePairs"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class AXIS_RESCALE_4(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Uint, "MaxNumberOfRescalePairs"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class AXIS_RESCALE_5(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (Uint, "MaxNumberOfRescalePairs"),
        (Indexorder, "IndexIncr"),
        (Addrtype, "Addressing"),
    ]


class BIT_MASK(Keyword):
    attrs = [(Ulong, "Mask")]  # mask to mask out single bits.


class BIT_OPERATION(Keyword):
    children = ["LEFT_SHIFT", "RIGHT_SHIFT", "SIGN_EXTEND"]
    block = True


class BLOB(Keyword):
    multiple = True
    block = True
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Ulong, "Address"),
        (Ulong, "Length"),
    ]
    children = [
        "CALIBRATION_ACCESS",
    ]


class BYTE_ORDER(Keyword):
    attrs = [(Byteorder, "ByteOrder")]


class CALIBRATION_ACCESS(Keyword):
    attrs = [
        (
            Enum,
            "Type",
            (
                "CALIBRATION",
                "NO_CALIBRATION",
                "NOT_IN_MCD_SYSTEM",
                "OFFLINE_CALIBRATION",
            ),
        )
    ]


class CALIBRATION_HANDLE(Keyword):
    multiple = True
    children = ["CALIBRATION_HANDLE_TEXT"]
    block = True
    attrs = [(Long, "Handle", MULTIPLE)]


class CALIBRATION_HANDLE_TEXT(Keyword):
    attrs = [(String, "Text")]


class CALIBRATION_METHOD(Keyword):
    multiple = True
    children = ["CALIBRATION_HANDLE"]
    block = True
    attrs = [
        (String, "Method"),  # the string identifies the calibration method to be used.
        # A convention regarding the meaning of the calibration
        # methods. The following strings are already in use: ‘
        # InCircuit’, ‘SERAM’, ‘DSERAP’, ‘BSERAP’
        (Ulong, "Version"),  # Version number of the method used.
    ]


class CHARACTERISTIC(Keyword):
    multiple = True
    block = True
    children = [
        "ANNOTATION",
        "AXIS_DESCR",
        "BIT_MASK",
        "BYTE_ORDER",
        "CALIBRATION_ACCESS",
        "COMPARISON_QUANTITY",
        "DEPENDENT_CHARACTERISTIC",
        "DISCRETE",
        "DISPLAY_IDENTIFIER",
        "ECU_ADDRESS_EXTENSION",
        "EXTENDED_LIMITS",
        "FORMAT",
        "FUNCTION_LIST",
        "GUARD_RAILS",
        "IF_DATA",
        "MAP_LIST",
        "MATRIX_DIM",
        "MAX_REFRESH",
        "NUMBER",
        "PHYS_UNIT",
        "READ_ONLY",
        "REF_MEMORY_SEGMENT",
        "STEP_SIZE",
        "SYMBOL_LINK",
        "VIRTUAL_CHARACTERISTIC",
    ]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (
            Enum,
            "Type",
            ("ASCII", "CURVE", "MAP", "CUBOID", "CUBE_4", "CUBE_5", "VAL_BLK", "VALUE"),
        ),
        (Ulong, "Address"),
        (Ident, "Deposit"),
        (Float, "MaxDiff"),
        (Ident, "Conversion"),
        (Float, "LowerLimit"),
        (Float, "UpperLimit"),
    ]


class COEFFS(Keyword):
    attrs = [
        (Float, "a"),
        (Float, "b"),
        (Float, "c"),
        (Float, "d"),
        (Float, "e"),
        (Float, "f"),
    ]


class COEFFS_LINEAR(Keyword):
    attrs = [
        (Float, "a"),
        (Float, "b"),
    ]


class COMPARISON_QUANTITY(Keyword):
    attrs = [(Ident, "Name")]


class COMPU_METHOD(Keyword):
    multiple = True
    block = True
    children = [
        "COEFFS",
        "COEFFS_LINEAR",
        "COMPU_TAB_REF",
        "FORMULA",
        "REF_UNIT",
        "STATUS_STRING_REF",
    ]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (
            Enum,
            "ConversionType",
            (
                "IDENTICAL",
                "FORM",
                "LINEAR",
                "RAT_FUNC",
                "TAB_INTP",
                "TAB_NOINTP",
                "TAB_VERB",
            ),
        ),
        (String, "Format"),
        (String, "Unit"),
    ]


class COMPU_TAB(Keyword):
    multiple = True
    block = True
    children = ["DEFAULT_VALUE", "DEFAULT_VALUE_NUMERIC"]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Enum, "ConversionType", ("TAB_INTP", "TAB_NOINTP")),
        (Uint, "NumberValuePairs"),
        ##
        ##  TODO: (float InVal float OutVal)* !!!
        ##
    ]


class COMPU_TAB_REF(Keyword):
    attrs = [(Ident, "ConversionTable")]


class COMPU_VTAB(Keyword):
    multiple = True
    block = True
    children = ["DEFAULT_VALUE"]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Enum, "ConversionType", ("TAB_VERB",)),
        (Uint, "NumberValuePairs"),
        ##
        ##  TODO: (float InVal float OutVal)* !!!
        ##
    ]


class COMPU_VTAB_RANGE(Keyword):
    multiple = True
    block = True
    children = ["DEFAULT_VALUE"]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Uint, "NumberValueTriples"),
    ]


class CPU_TYPE(Keyword):
    attrs = [(String, "CPU")]


class CURVE_AXIS_REF(Keyword):
    attrs = [(Ident, "CurveAxis")]


class CUSTOMER(Keyword):
    attrs = [(String, "Customer")]


class CUSTOMER_NO(Keyword):
    attrs = [(String, "Number")]


class DATA_SIZE(Keyword):
    attrs = [(Uint, "Size")]


class DEF_CHARACTERISTIC(Keyword):
    block = True
    attrs = [(Ident, "Identifier", MULTIPLE)]


class DEFAULT_VALUE(Keyword):
    attrs = [(String, "display_string")]


class DEFAULT_VALUE_NUMERIC(Keyword):
    attrs = [(Float, "display_value")]


class DEPENDENT_CHARACTERISTIC(Keyword):
    block = True
    attrs = [
        (String, "Formula"),
        (Ident, "Characteristic", MULTIPLE),
    ]


class DEPOSIT(Keyword):
    attrs = [(Enum, "Mode", ("ABSOLUTE", "DIFFERENCE"))]


class DISCRETE(Keyword):
    pass


class DISPLAY_IDENTIFIER(Keyword):
    attrs = [(Ident, "display_name")]


class DIST_OP_X(Keyword):
    attrs = [(Uint, "Position"), (Datatype, "Datatype")]


class DIST_OP_Y(Keyword):
    attrs = [(Uint, "Position"), (Datatype, "Datatype")]


class DIST_OP_Z(Keyword):
    attrs = [(Uint, "Position"), (Datatype, "Datatype")]


class DIST_OP_4(Keyword):
    attrs = [(Uint, "Position"), (Datatype, "Datatype")]


class DIST_OP_5(Keyword):
    attrs = [(Uint, "Position"), (Datatype, "Datatype")]


class ECU(Keyword):
    attrs = [(String, "ControlUnit")]


class ECU_ADDRESS(Keyword):
    attrs = [(Ulong, "Address")]


class ECU_ADDRESS_EXTENSION(Keyword):
    attrs = [(Int, "Extension")]


class ECU_CALIBRATION_OFFSET(Keyword):
    attrs = [(Long, "Offset")]


class EPK(Keyword):
    attrs = [(String, "Identifier")]


class ERROR_MASK(Keyword):
    attrs = [(Ulong, "Mask")]


class EXTENDED_LIMITS(Keyword):
    attrs = [
        (Float, "LowerLimit"),
        (Float, "UpperLimit"),
    ]


class FIX_AXIS_PAR(Keyword):
    attrs = [
        (Int, "Offset"),
        (Int, "Shift"),
        (Uint, "Numberapo"),
    ]


class FIX_AXIS_PAR_DIST(Keyword):
    attrs = [
        (Int, "Offset"),
        (Int, "Distance"),
        (Uint, "Numberapo"),
    ]


class FIX_AXIS_PAR_LIST(Keyword):
    block = True
    attrs = [
        (Float, "AxisPts_Value", MULTIPLE),
    ]


class FIX_NO_AXIS_PTS_X(Keyword):
    attrs = [(Uint, "NumberOfAxisPoints")]


class FIX_NO_AXIS_PTS_Y(Keyword):
    attrs = [(Uint, "NumberOfAxisPoints")]


class FIX_NO_AXIS_PTS_Z(Keyword):
    attrs = [(Uint, "NumberOfAxisPoints")]


class FIX_NO_AXIS_PTS_4(Keyword):
    attrs = [(Uint, "NumberOfAxisPoints")]


class FIX_NO_AXIS_PTS_5(Keyword):
    attrs = [(Uint, "NumberOfAxisPoints")]


class FNC_VALUES(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
        (
            Enum,
            "IndexMode",
            (
                "ALTERNATE_CURVES",
                "ALTERNATE_WITH_X",
                "ALTERNATE_WITH_Y",
                "COLUMN_DIR",
                "ROW_DIR",
            ),
        ),
        (Addrtype, "Addresstype"),
    ]


class FORMAT(Keyword):
    attrs = [(String, "FormatString")]


class FORMULA(Keyword):
    children = ["FORMULA_INV"]
    block = True
    attrs = [(String, "F_x")]


class FORMULA_INV(Keyword):
    attrs = [(String, "G_x")]


class FRAME(Keyword):
    children = ["FRAME_MEASUREMENT", "IF_DATA"]
    block = True
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Uint, "ScalingUnit"),
        (Ulong, "Rate"),
    ]


class FRAME_MEASUREMENT(Keyword):
    attrs = [(Ident, "Identifier", MULTIPLE)]


class FUNCTION(Keyword):
    multiple = True
    block = True
    children = [
        "ANNOTATION",
        "DEF_CHARACTERISTIC",
        "FUNCTION_VERSION",
        "IF_DATA",
        "IN_MEASUREMENT",
        "LOC_MEASUREMENT",
        "OUT_MEASUREMENT",
        "REF_CHARACTERISTIC",
        "SUB_FUNCTION",
    ]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
    ]


class FUNCTION_LIST(Keyword):
    block = True
    attrs = [
        (Ident, "Name", MULTIPLE),
    ]


class FUNCTION_VERSION(Keyword):
    attrs = [(String, "VersionIdentifier")]


class GROUP(Keyword):
    multiple = True
    children = [
        "ANNOTATION",
        "FUNCTION_LIST",
        "IF_DATA",
        "REF_CHARACTERISTIC",
        "REF_MEASUREMENT",
        "ROOT",
        "SUB_GROUP",
    ]
    block = True
    attrs = [
        (Ident, "GroupName"),
        (String, "GroupLongIdentifier"),
    ]


class GUARD_RAILS(Keyword):
    pass


class HEADER(Keyword):
    children = ["PROJECT_NO", "VERSION"]
    block = True
    attrs = [(String, "Comment")]


class IDENTIFICATION(Keyword):
    attrs = [(Uint, "Position"), (Datatype, "Datatype")]


class IF_DATA(Keyword):
    multiple = True
    block = True
    attrs = [
        (String, "raw"),
    ]
    """
        Data record to describe interface specific data. The parameters associated with this
        keyword have to be described in the ASAM MCD-2MC metalanguage.
        These parameters describe e.g. the access methods to the measurement data collection,
        serial communication and so on.
    """


class IN_MEASUREMENT(Keyword):
    block = True
    attrs = [(Ident, "Identifier", MULTIPLE)]


class INSTANCE(Keyword):
    block = True
    multiple = True
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Ident, "TypeName"),
        (Ulong, "Address"),
    ]
    children = [
        "COMPARISON_QUANTITY",
        "DEPENDENT_CHARACTERISTIC",
        "DISPLAY_IDENTIFIER",
        "ECU_ADDRESS_EXTENSION",
        "IF_DATA",
        "NUMBER",
        "MATRIX_DIM",
        "SYMBOL_LINK",
    ]


class LAYOUT(Keyword):
    attrs = [(Enum, "IndexMode", ("ROW_DIR", "COLUMN_DIR"))]


class LEFT_SHIFT(Keyword):
    attrs = [(Ulong, "Bitcount")]


class LOC_MEASUREMENT(Keyword):
    block = True
    attrs = [(Ident, "Identifier", MULTIPLE)]


class MAP_LIST(Keyword):
    block = True
    attrs = [(Ident, "Name", MULTIPLE)]


class MODEL_LINK(Keyword):
    attrs = [
        (String, "Link"),
    ]


class MATRIX_DIM(Keyword):
    attrs = [(Uint, "Numbers", MULTIPLE)]


class MAX_GRAD(Keyword):
    attrs = [(Float, "MaxGradient")]


class MAX_REFRESH(Keyword):
    attrs = [
        (Uint, "ScalingUnit"),
        (Ulong, "Rate"),
    ]


class MEASUREMENT(Keyword):
    multiple = True
    block = True
    children = [
        "ANNOTATION",
        "ARRAY_SIZE",
        "BIT_MASK",
        "BIT_OPERATION",
        "BYTE_ORDER",
        "DISCRETE",
        "DISPLAY_IDENTIFIER",
        "ECU_ADDRESS",
        "ECU_ADDRESS_EXTENSION",
        "ERROR_MASK",
        "FORMAT",
        "FUNCTION_LIST",
        "IF_DATA",
        "LAYOUT",
        "MATRIX_DIM",
        "MAX_REFRESH",
        "PHYS_UNIT",
        "READ_WRITE",
        "REF_MEMORY_SEGMENT",
        "SYMBOL_LINK",
        "VIRTUAL",
    ]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Datatype, "Datatype"),
        (Ident, "Conversion"),
        (Uint, "Resolution"),
        (Float, "Accuracy"),
        (Float, "LowerLimit"),
        (Float, "UpperLimit"),
    ]


class MEMORY_LAYOUT(Keyword):
    multiple = True
    children = ["IF_DATA"]
    block = True
    attrs = [
        (Enum, "PrgType", ("PRG_CODE", "PRG_DATA", "PRG_RESERVED")),
        (Ulong, "Address"),
        (Ulong, "Size"),
        (Long, "Offset_0"),
        (Long, "Offset_1"),
        (Long, "Offset_2"),
        (Long, "Offset_3"),
        (Long, "Offset_4"),
    ]


class MEMORY_SEGMENT(Keyword):
    multiple = True
    children = ["IF_DATA"]
    block = True
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (
            Enum,
            "PrgType",
            (
                "CALIBRATION_VARIABLES",
                "CODE",
                "DATA",
                "EXCLUDE_FROM_FLASH",
                "OFFLINE_DATA",
                "RESERVED",
                "SERAM",
                "VARIABLES",
            ),
        ),
        (Enum, "MemoryType", ("EEPROM", "EPROM", "FLASH", "RAM", "ROM", "REGISTER", "NOT_IN_ECU")),
        (Enum, "Attribute", ("INTERN", "EXTERN")),
        (Ulong, "Address"),
        (Ulong, "Size"),
        (Long, "Offset_0"),
        (Long, "Offset_1"),
        (Long, "Offset_2"),
        (Long, "Offset_3"),
        (Long, "Offset_4"),
    ]


class MOD_COMMON(Keyword):
    block = True
    children = [
        "ALIGNMENT_BYTE",
        "ALIGNMENT_FLOAT16_IEEE",
        "ALIGNMENT_FLOAT32_IEEE",
        "ALIGNMENT_FLOAT64_IEEE",
        "ALIGNMENT_INT64",
        "ALIGNMENT_LONG",
        "ALIGNMENT_WORD",
        "BYTE_ORDER",
        "DATA_SIZE",
        "DEPOSIT",
        "S_REC_LAYOUT",
    ]
    attrs = [
        (String, "Comment"),
    ]


class MOD_PAR(Keyword):
    block = True
    children = [
        "ADDR_EPK",
        "CALIBRATION_METHOD",
        "CPU_TYPE",
        "CUSTOMER",
        "CUSTOMER_NO",
        "ECU",
        "ECU_CALIBRATION_OFFSET",
        "EPK",
        "MEMORY_LAYOUT",
        "MEMORY_SEGMENT",
        "NO_OF_INTERFACES",
        "PHONE_NO",
        "SUPPLIER",
        "SYSTEM_CONSTANT",
        "USER",
        "VERSION",
    ]
    attrs = [
        (String, "Comment"),
    ]


class MODULE(Keyword):
    multiple = True
    block = True
    children = [
        "A2ML",
        "AXIS_PTS",
        "CHARACTERISTIC",
        "COMPU_METHOD",
        "COMPU_TAB",
        "COMPU_VTAB",
        "COMPU_VTAB_RANGE",
        "FRAME",
        "FUNCTION",
        "GROUP",
        "IF_DATA",
        "INSTANCE",
        "MEASUREMENT",
        "MOD_COMMON",
        "MOD_PAR",
        "RECORD_LAYOUT",
        "TYPEDEF_CHARACTERISTIC",
        "TYPEDEF_MEASUREMENT",
        "TYPEDEF_STRUCTURE",
        "UNIT",
        "USER_RIGHTS",
        "VARIANT_CODING",
    ]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
    ]


class MONOTONY(Keyword):
    attrs = [
        (
            Enum,
            "Monotony",
            (
                "MON_DECREASE",
                "MON_INCREASE",
                "STRICT_DECREASE",
                "STRICT_INCREASE",
                "MONOTONOUS",
                "STRICT_MON",
                "NOT_MON",
            ),
        )
    ]


class NO_AXIS_PTS_X(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NO_AXIS_PTS_Y(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NO_AXIS_PTS_Z(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NO_AXIS_PTS_4(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NO_AXIS_PTS_5(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NO_OF_INTERFACES(Keyword):
    attrs = [
        (Uint, "Num"),
    ]


class NO_RESCALE_X(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NO_RESCALE_Y(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NO_RESCALE_Z(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NO_RESCALE_4(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NO_RESCALE_5(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class NUMBER(Keyword):
    attrs = [(Uint, "Number")]


class OFFSET_X(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class OFFSET_Y(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class OFFSET_Z(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class OFFSET_4(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class OFFSET_5(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class OUT_MEASUREMENT(Keyword):
    block = True
    attrs = [
        (Ident, "Identifier", MULTIPLE),
    ]


class PHONE_NO(Keyword):
    attrs = [
        (String, "Telnum"),
    ]


class PHYS_UNIT(Keyword):
    attrs = [
        (String, "Unit"),
    ]


class PROJECT(Keyword):
    children = ["HEADER", "MODULE"]
    block = True
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
    ]


class PROJECT_NO(Keyword):
    attrs = [
        (Ident, "ProjectNumber"),
    ]


class READ_ONLY(Keyword):
    pass


class READ_WRITE(Keyword):
    pass


class RECORD_LAYOUT(Keyword):
    multiple = True
    block = True
    children = [
        "ALIGNMENT_BYTE",
        "ALIGNMENT_FLOAT16_IEEE",
        "ALIGNMENT_FLOAT32_IEEE",
        "ALIGNMENT_FLOAT64_IEEE",
        "ALIGNMENT_INT64",
        "ALIGNMENT_LONG",
        "ALIGNMENT_WORD",
        "AXIS_PTS_X",
        "AXIS_PTS_Y",
        "AXIS_PTS_Z",
        "AXIS_PTS_4",
        "AXIS_PTS_5",
        "AXIS_RESCALE_X",
        "AXIS_RESCALE_Y",
        "AXIS_RESCALE_Z",
        "AXIS_RESCALE_4",
        "AXIS_RESCALE_5",
        "DIST_OP_X",
        "DIST_OP_Y",
        "DIST_OP_Z",
        "DIST_OP_4",
        "DIST_OP_5",
        "FIX_NO_AXIS_PTS_X",
        "FIX_NO_AXIS_PTS_Y",
        "FIX_NO_AXIS_PTS_Z",
        "FIX_NO_AXIS_PTS_4",
        "FIX_NO_AXIS_PTS_5",
        "FNC_VALUES",
        "IDENTIFICATION",
        "NO_AXIS_PTS_X",
        "NO_AXIS_PTS_Y",
        "NO_AXIS_PTS_Z",
        "NO_AXIS_PTS_4",
        "NO_AXIS_PTS_5",
        "STATIC_RECORD_LAYOUT",
        "NO_RESCALE_X",
        "NO_RESCALE_Y",
        "NO_RESCALE_Z",
        "NO_RESCALE_4",
        "NO_RESCALE_5",
        "OFFSET_X",
        "OFFSET_Y",
        "OFFSET_Z",
        "OFFSET_4",
        "OFFSET_5",
        "RESERVED",
        "RIP_ADDR_W",
        "RIP_ADDR_X",
        "RIP_ADDR_Y",
        "RIP_ADDR_Z",
        "RIP_ADDR_4",
        "RIP_ADDR_5",
        "SHIFT_OP_X",
        "SHIFT_OP_Y",
        "SHIFT_OP_Z",
        "SHIFT_OP_4",
        "SHIFT_OP_5",
        "SRC_ADDR_X",
        "SRC_ADDR_Y",
        "SRC_ADDR_Z",
        "SRC_ADDR_4",
        "SRC_ADDR_5",
    ]
    attrs = [
        (Ident, "Name"),
    ]


class TRANSFORMER_IN_OBJECTS(Keyword):
    block = True
    attrs = [
        (Ident, "Identifier", MULTIPLE),
    ]


class TRANSFORMER_OUT_OBJECTS(Keyword):
    block = True
    attrs = [
        (Ident, "Identifier", MULTIPLE),
    ]


class REF_CHARACTERISTIC(Keyword):
    block = True
    attrs = [
        (Ident, "Identifier", MULTIPLE),
    ]


class REF_GROUP(Keyword):
    multiple = True
    block = True
    attrs = [
        (Ident, "Identifier", MULTIPLE),
    ]


class REF_MEASUREMENT(Keyword):
    block = True
    attrs = [
        (Ident, "Identifier", MULTIPLE),
    ]


class REF_MEMORY_SEGMENT(Keyword):
    attrs = [
        (Ident, "Name"),
    ]


class REF_UNIT(Keyword):
    attrs = [
        (Ident, "Unit"),
    ]


class RESERVED(Keyword):
    multiple = True
    attrs = [
        (Uint, "Position"),
        (Datasize, "DataSize"),
    ]


class RIGHT_SHIFT(Keyword):
    attrs = [
        (Ulong, "Bitcount"),
    ]


class RIP_ADDR_W(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class RIP_ADDR_X(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class RIP_ADDR_Y(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class RIP_ADDR_Z(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class RIP_ADDR_4(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class RIP_ADDR_5(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class ROOT(Keyword):
    pass


class SHIFT_OP_X(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class SHIFT_OP_Y(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class SHIFT_OP_Z(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class SHIFT_OP_4(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class SHIFT_OP_5(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class SIGN_EXTEND(Keyword):
    pass


class SI_EXPONENTS(Keyword):
    attrs = [
        (Int, "Length"),
        (Int, "Mass"),
        (Int, "Time"),
        (Int, "ElectricCurrent"),
        (Int, "Temperature"),
        (Int, "AmountOfSubstance"),
        (Int, "LuminousIntensity"),
    ]


class SRC_ADDR_X(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class SRC_ADDR_Y(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class SRC_ADDR_Z(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class SRC_ADDR_4(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class SRC_ADDR_5(Keyword):
    attrs = [
        (Uint, "Position"),
        (Datatype, "Datatype"),
    ]


class STATIC_RECORD_LAYOUT(Keyword):
    pass


class STATUS_STRING_REF(Keyword):
    attrs = [
        (Ident, "ConversionTable"),
    ]


class STEP_SIZE(Keyword):
    attrs = [
        (Float, "StepSize"),
    ]


class SUB_FUNCTION(Keyword):
    block = True
    attrs = [
        (Ident, "Identifier", MULTIPLE),
    ]


class SUB_GROUP(Keyword):
    block = True
    attrs = [
        (Ident, "Identifier", MULTIPLE),
    ]


class SUPPLIER(Keyword):
    attrs = [
        (String, "Manufacturer"),
    ]


class SYMBOL_LINK(Keyword):
    attrs = [
        (String, "SymbolName"),
        (Long, "Offset"),
    ]


class SYMBOL_TYPE_LINK(Keyword):
    attrs = [
        (String, "Link"),
    ]


class SYSTEM_CONSTANT(Keyword):
    multiple = True
    attrs = [
        (String, "Name"),
        (String, "Value"),
    ]


class S_REC_LAYOUT(Keyword):
    attrs = [
        (Ident, "Name"),
    ]


class STRUCTURE_COMPONENT(Keyword):
    multiple = True
    block = True
    attrs = [
        (Ident, "Name"),
        (Ident, "Type_Ref"),
        (Ulong, "Offset"),
    ]
    children = [
        "MATRIX_DIM",
        "NUMBER",
        "SYMBOL_TYPE_LINK",
    ]


class TRANSFORMER(Keyword):
    multiple = True
    block = True

    attrs = [
        (Ident, "Name"),
        (String, "Version"),
        (String, "DllName32"),
        (String, "DllName64"),
        (Ulong, "Timeout"),
        (Ident, "Trigger"),
        (Ident, "Reverse"),
    ]

    children = [
        "TRANSFORMER_IN_OBJECTS",
        "TRANSFORMER_OUT_OBJECTS",
    ]


class TYPEDEF_AXIS(Keyword):
    multiple = True
    block = True
    children = [
        "ANNOTATION",
        "BYTE_ORDER",
        "CALIBRATION_ACCESS",
        "DEPOSIT",
        "EXTENDED_LIMITS",
        "FORMAT",
        "GUARD_RAILS",
        "MONOTONY",
        "PHYS_UNIT",
        "READ_ONLY",
        "REF_MEMORY_SEGMENT",
        "STEP_SIZE",
    ]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Ident, "InputQuantity"),
        (Ident, "DepositAttr"),
        (Float, "MaxDiff"),
        (Ident, "Conversion"),
        (Uint, "MaxAxisPoints"),
        (Float, "LowerLimit"),
        (Float, "UpperLimit"),
    ]


class TYPEDEF_CHARACTERISTIC(Keyword):
    multiple = True
    block = True
    children = [
        "ANNOTATION",
        "AXIS_DESCR",
        "BIT_MASK",
        "BYTE_ORDER",
        "CALIBRATION_ACCESS",
        "DISCRETE",
        "EXTENDED_LIMITS",
        "FORMAT",
        "GUARD_RAILS",
        "MAX_REFRESH",
        "NUMBER",
        "PHYS_UNIT",
        "READ_ONLY",
        "REF_MEMORY_SEGMENT",
        "STEP_SIZE",
        "VIRTUAL_CHARACTERISTIC",
    ]
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (
            Enum,
            "Type",
            ("ASCII", "CURVE", "MAP", "CUBOID", "CUBE_4", "CUBE_5", "VAL_BLK", "VALUE"),
        ),
        (Ident, "Deposit"),
        (Float, "MaxDiff"),
        (Ident, "Conversion"),
        (Float, "LowerLimit"),
        (Float, "UpperLimit"),
    ]


class TYPEDEF_MEASUREMENT(Keyword):
    multiple = True
    block = True
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Datatype, "Datatype"),
        (Ident, "Conversion"),
        (Uint, "Resolution"),
        (Float, "Accuracy"),
        (Float, "LowerLimit"),
        (Float, "UpperLimit"),
    ]


class TYPEDEF_STRUCTURE(Keyword):
    multiple = True
    block = True
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Ulong, "Size"),
    ]
    children = [
        "STRUCTURE_COMPONENT",
        "SYMBOL_TYPE_LINK",
    ]


class UNIT(Keyword):
    multiple = True
    children = ["SI_EXPONENTS", "REF_UNIT", "UNIT_CONVERSION"]
    block = True
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (String, "Display"),
        (Enum, "Type", ("DERIVED", "EXTENDED_SI")),
    ]


class UNIT_CONVERSION(Keyword):
    attrs = [
        (Float, "Gradient"),
        (Float, "Offset"),
    ]


class USER(Keyword):
    attrs = [(String, "UserName")]


class USER_RIGHTS(Keyword):
    multiple = True
    children = ["READ_ONLY", "REF_GROUP"]
    block = True
    attrs = [(Ident, "UserLevelId")]


class VAR_ADDRESS(Keyword):
    block = True
    attrs = [
        (Ulong, "Address", MULTIPLE),
    ]


class VAR_CHARACTERISTIC(Keyword):
    multiple = True
    children = ["VAR_ADDRESS"]
    block = True
    attrs = [
        (Ident, "Name"),
        (Ident, "CriterionName", MULTIPLE),
    ]


class VAR_CRITERION(Keyword):
    multiple = True
    children = ["VAR_MEASUREMENT", "VAR_SELECTION_CHARACTERISTIC"]
    block = True
    attrs = [
        (Ident, "Name"),
        (String, "LongIdentifier"),
        (Ident, "Value", MULTIPLE),
    ]


class VAR_FORBIDDEN_COMB(Keyword):
    multiple = True
    block = True
    # attrs = [
    #    (Ident, "CriterionName"),
    #    (Ident, "CriterionValue")
    #    ## Hinweis: die Multiplizität bezieht sich hier auf das Tupel!!!
    # ]


class VAR_MEASUREMENT(Keyword):
    attrs = [
        (Ident, "Name"),
    ]


class VAR_NAMING(Keyword):
    attrs = [(Enum, "Tag", ("NUMERIC", "APLHA"))]


class VAR_SELECTION_CHARACTERISTIC(Keyword):
    attrs = [
        (Ident, "Name"),
    ]


class VAR_SEPARATOR(Keyword):
    attrs = [
        (String, "Separator"),
    ]


class VARIANT_CODING(Keyword):
    children = [
        "VAR_CHARACTERISTIC",
        "VAR_CRITERION",
        "VAR_FORBIDDEN_COMB",
        "VAR_NAMING",
        "VAR_SEPARATOR",
    ]
    block = True


class VERSION(Keyword):
    attrs = [(String, "VersionIdentifier")]


class VIRTUAL(Keyword):
    block = True
    attrs = [
        (Ident, "MeasuringChannel", MULTIPLE),
    ]


class VIRTUAL_CHARACTERISTIC(Keyword):
    block = True
    attrs = [
        (String, "Formula"),
        (Ident, "Characteristic", MULTIPLE),
    ]


##
##
##
class RootElement(Keyword):
    children = ["ASAP2_VERSION", "A2ML_VERSION", "PROJECT"]


###
###
###

KEYWORD_MAP = {
    "A2ML": A2ML,
    "A2ML_VERSION": A2ML_VERSION,
    "ADDR_EPK": ADDR_EPK,
    "ALIGNMENT_BYTE": ALIGNMENT_BYTE,
    "ALIGNMENT_FLOAT16_IEEE": ALIGNMENT_FLOAT16_IEEE,
    "ALIGNMENT_FLOAT32_IEEE": ALIGNMENT_FLOAT32_IEEE,
    "ALIGNMENT_FLOAT64_IEEE": ALIGNMENT_FLOAT64_IEEE,
    "ALIGNMENT_INT64": ALIGNMENT_INT64,
    "ALIGNMENT_LONG": ALIGNMENT_LONG,
    "ALIGNMENT_WORD": ALIGNMENT_WORD,
    "ANNOTATION": ANNOTATION,
    "ANNOTATION_LABEL": ANNOTATION_LABEL,
    "ANNOTATION_ORIGIN": ANNOTATION_ORIGIN,
    "ANNOTATION_TEXT": ANNOTATION_TEXT,
    "ARRAY_SIZE": ARRAY_SIZE,
    "ASAP2_VERSION": ASAP2_VERSION,
    "AXIS_DESCR": AXIS_DESCR,
    "AXIS_PTS": AXIS_PTS,
    "AXIS_PTS_REF": AXIS_PTS_REF,
    "AXIS_PTS_X": AXIS_PTS_X,
    "AXIS_PTS_Y": AXIS_PTS_Y,
    "AXIS_PTS_Z": AXIS_PTS_Z,
    "AXIS_PTS_4": AXIS_PTS_4,
    "AXIS_PTS_5": AXIS_PTS_5,
    "AXIS_RESCALE_X": AXIS_RESCALE_X,
    "AXIS_RESCALE_Y": AXIS_RESCALE_Y,
    "AXIS_RESCALE_Z": AXIS_RESCALE_Z,
    "AXIS_RESCALE_4": AXIS_RESCALE_4,
    "AXIS_RESCALE_5": AXIS_RESCALE_5,
    "BIT_MASK": BIT_MASK,
    "BIT_OPERATION": BIT_OPERATION,
    "BLOB": BLOB,
    "BYTE_ORDER": BYTE_ORDER,
    "CALIBRATION_ACCESS": CALIBRATION_ACCESS,
    "CALIBRATION_HANDLE": CALIBRATION_HANDLE,
    "CALIBRATION_HANDLE_TEXT": CALIBRATION_HANDLE_TEXT,
    "CALIBRATION_METHOD": CALIBRATION_METHOD,
    "CHARACTERISTIC": CHARACTERISTIC,
    "COEFFS": COEFFS,
    "COEFFS_LINEAR": COEFFS_LINEAR,
    "COMPARISON_QUANTITY": COMPARISON_QUANTITY,
    "COMPU_METHOD": COMPU_METHOD,
    "COMPU_TAB": COMPU_TAB,
    "COMPU_TAB_REF": COMPU_TAB_REF,
    "COMPU_VTAB": COMPU_VTAB,
    "COMPU_VTAB_RANGE": COMPU_VTAB_RANGE,
    "CPU_TYPE": CPU_TYPE,
    "CURVE_AXIS_REF": CURVE_AXIS_REF,
    "CUSTOMER": CUSTOMER,
    "CUSTOMER_NO": CUSTOMER_NO,
    "DATA_SIZE": DATA_SIZE,
    "DEF_CHARACTERISTIC": DEF_CHARACTERISTIC,
    "DEFAULT_VALUE": DEFAULT_VALUE,
    "DEFAULT_VALUE_NUMERIC": DEFAULT_VALUE_NUMERIC,
    "DEPENDENT_CHARACTERISTIC": DEPENDENT_CHARACTERISTIC,
    "DEPOSIT": DEPOSIT,
    "DISCRETE": DISCRETE,
    "DISPLAY_IDENTIFIER": DISPLAY_IDENTIFIER,
    "DIST_OP_X": DIST_OP_X,
    "DIST_OP_Y": DIST_OP_Y,
    "DIST_OP_Z": DIST_OP_Z,
    "DIST_OP_4": DIST_OP_4,
    "DIST_OP_5": DIST_OP_5,
    "ECU": ECU,
    "ECU_ADDRESS": ECU_ADDRESS,
    "ECU_ADDRESS_EXTENSION": ECU_ADDRESS_EXTENSION,
    "ECU_CALIBRATION_OFFSET": ECU_CALIBRATION_OFFSET,
    "EPK": EPK,
    "ERROR_MASK": ERROR_MASK,
    "EXTENDED_LIMITS": EXTENDED_LIMITS,
    "FIX_AXIS_PAR": FIX_AXIS_PAR,
    "FIX_AXIS_PAR_DIST": FIX_AXIS_PAR_DIST,
    "FIX_AXIS_PAR_LIST": FIX_AXIS_PAR_LIST,
    "FIX_NO_AXIS_PTS_X": FIX_NO_AXIS_PTS_X,
    "FIX_NO_AXIS_PTS_Y": FIX_NO_AXIS_PTS_Y,
    "FIX_NO_AXIS_PTS_Z": FIX_NO_AXIS_PTS_Z,
    "FIX_NO_AXIS_PTS_4": FIX_NO_AXIS_PTS_4,
    "FIX_NO_AXIS_PTS_5": FIX_NO_AXIS_PTS_5,
    "FNC_VALUES": FNC_VALUES,
    "FORMAT": FORMAT,
    "FORMULA": FORMULA,
    "FORMULA_INV": FORMULA_INV,
    "FRAME": FRAME,
    "FRAME_MEASUREMENT": FRAME_MEASUREMENT,
    "FUNCTION": FUNCTION,
    "FUNCTION_LIST": FUNCTION_LIST,
    "FUNCTION_VERSION": FUNCTION_VERSION,
    "GROUP": GROUP,
    "GUARD_RAILS": GUARD_RAILS,
    "HEADER": HEADER,
    "IDENTIFICATION": IDENTIFICATION,
    "IF_DATA": IF_DATA,
    "IN_MEASUREMENT": IN_MEASUREMENT,
    "INSTANCE": INSTANCE,
    "LAYOUT": LAYOUT,
    "LEFT_SHIFT": LEFT_SHIFT,
    "LOC_MEASUREMENT": LOC_MEASUREMENT,
    "MAP_LIST": MAP_LIST,
    "MATRIX_DIM": MATRIX_DIM,
    "MAX_GRAD": MAX_GRAD,
    "MAX_REFRESH": MAX_REFRESH,
    "MEASUREMENT": MEASUREMENT,
    "MEMORY_LAYOUT": MEMORY_LAYOUT,
    "MEMORY_SEGMENT": MEMORY_SEGMENT,
    "MOD_COMMON": MOD_COMMON,
    "MOD_PAR": MOD_PAR,
    "MODULE": MODULE,
    "MODEL_LINK": MODEL_LINK,
    "MONOTONY": MONOTONY,
    "NO_AXIS_PTS_X": NO_AXIS_PTS_X,
    "NO_AXIS_PTS_Y": NO_AXIS_PTS_Y,
    "NO_AXIS_PTS_Z": NO_AXIS_PTS_Z,
    "NO_AXIS_PTS_4": NO_AXIS_PTS_4,
    "NO_AXIS_PTS_5": NO_AXIS_PTS_5,
    "NO_OF_INTERFACES": NO_OF_INTERFACES,
    "NO_RESCALE_X": NO_RESCALE_X,
    "NO_RESCALE_Y": NO_RESCALE_Y,
    "NO_RESCALE_Z": NO_RESCALE_Z,
    "NO_RESCALE_4": NO_RESCALE_4,
    "NO_RESCALE_5": NO_RESCALE_5,
    "NUMBER": NUMBER,
    "OFFSET_X": OFFSET_X,
    "OFFSET_Y": OFFSET_Y,
    "OFFSET_Z": OFFSET_Z,
    "OFFSET_4": OFFSET_4,
    "OFFSET_5": OFFSET_5,
    "OUT_MEASUREMENT": OUT_MEASUREMENT,
    "PHONE_NO": PHONE_NO,
    "PHYS_UNIT": PHYS_UNIT,
    "PROJECT": PROJECT,
    "PROJECT_NO": PROJECT_NO,
    "READ_ONLY": READ_ONLY,
    "READ_WRITE": READ_WRITE,
    "RECORD_LAYOUT": RECORD_LAYOUT,
    "REF_CHARACTERISTIC": REF_CHARACTERISTIC,
    "REF_GROUP": REF_GROUP,
    "REF_MEASUREMENT": REF_MEASUREMENT,
    "REF_MEMORY_SEGMENT": REF_MEMORY_SEGMENT,
    "REF_UNIT": REF_UNIT,
    "RESERVED": RESERVED,
    "RIGHT_SHIFT": RIGHT_SHIFT,
    "RIP_ADDR_W": RIP_ADDR_W,
    "RIP_ADDR_X": RIP_ADDR_X,
    "RIP_ADDR_Y": RIP_ADDR_Y,
    "RIP_ADDR_Z": RIP_ADDR_Z,
    "RIP_ADDR_4": RIP_ADDR_4,
    "RIP_ADDR_5": RIP_ADDR_5,
    "ROOT": ROOT,
    "SHIFT_OP_X": SHIFT_OP_X,
    "SHIFT_OP_Y": SHIFT_OP_Y,
    "SHIFT_OP_Z": SHIFT_OP_Z,
    "SHIFT_OP_4": SHIFT_OP_4,
    "SHIFT_OP_5": SHIFT_OP_5,
    "SIGN_EXTEND": SIGN_EXTEND,
    "SI_EXPONENTS": SI_EXPONENTS,
    "SRC_ADDR_X": SRC_ADDR_X,
    "SRC_ADDR_Y": SRC_ADDR_Y,
    "SRC_ADDR_Z": SRC_ADDR_Z,
    "SRC_ADDR_4": SRC_ADDR_4,
    "SRC_ADDR_5": SRC_ADDR_5,
    "STATIC_RECORD_LAYOUT": STATIC_RECORD_LAYOUT,
    "STATUS_STRING_REF": STATUS_STRING_REF,
    "STEP_SIZE": STEP_SIZE,
    "STRUCTURE_COMPONENT": STRUCTURE_COMPONENT,
    "SUB_FUNCTION": SUB_FUNCTION,
    "SUB_GROUP": SUB_GROUP,
    "SUPPLIER": SUPPLIER,
    "SYMBOL_LINK": SYMBOL_LINK,
    "SYMBOL_TYPE_LINK": SYMBOL_TYPE_LINK,
    "SYSTEM_CONSTANT": SYSTEM_CONSTANT,
    "S_REC_LAYOUT": S_REC_LAYOUT,
    "TRANSFORMER_IN_OBJECTS": TRANSFORMER_IN_OBJECTS,
    "TRANSFORMER_OUT_OBJECTS": TRANSFORMER_OUT_OBJECTS,
    "TRANSFORMER": TRANSFORMER,
    "TYPEDEF_AXIS": TYPEDEF_AXIS,
    "TYPEDEF_CHARACTERISTIC": TYPEDEF_CHARACTERISTIC,
    "TYPEDEF_MEASUREMENT": TYPEDEF_MEASUREMENT,
    "TYPEDEF_STRUCTURE": TYPEDEF_STRUCTURE,
    "UNIT": UNIT,
    "UNIT_CONVERSION": UNIT_CONVERSION,
    "USER": USER,
    "USER_RIGHTS": USER_RIGHTS,
    "VAR_ADDRESS": VAR_ADDRESS,
    "VAR_CHARACTERISTIC": VAR_CHARACTERISTIC,
    "VAR_CRITERION": VAR_CRITERION,
    "VAR_FORBIDDEN_COMB": VAR_FORBIDDEN_COMB,
    "VAR_MEASUREMENT": VAR_MEASUREMENT,
    "VAR_NAMING": VAR_NAMING,
    "VAR_SELECTION_CHARACTERISTIC": VAR_SELECTION_CHARACTERISTIC,
    "VAR_SEPARATOR": VAR_SEPARATOR,
    "VARIANT_CODING": VARIANT_CODING,
    "VERSION": VERSION,
    "VIRTUAL": VIRTUAL,
    "VIRTUAL_CHARACTERISTIC": VIRTUAL_CHARACTERISTIC,
}


class A2LElement:
    def __str__(self):
        result = []
        result.append("%s {" % self.__class__.__name__)
        for attr in self.attrs:
            value = getattr(self, attr)
            formatStr = '    %s = "%s";' if isinstance(value, str) else "    %s = %s;"
            result.append(formatStr % (attr, value))
        result.append("}")

        return "\n".join(result)


def instanceFactory(className, **kws):
    """Create an instance of a given class."""
    klass = type(str(className), (A2LElement,), {})
    inst = klass()
    inst.attrs = []
    for k, v in kws.items():
        setattr(inst, k, v)
        inst.attrs.append(k)
    inst.children = []
    return inst



#!/usr/bin/env python

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2023 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"

import importlib
import sys
from collections import namedtuple

import antlr4
from antlr4.error.ErrorListener import ErrorListener

from pya2l import model


# from antlr4.atn.ParserATNSimulator import ParserATNSimulator


ResultType = namedtuple("ResultType", "db listener_result")


class MyErrorListener(ErrorListener):
    """Minimal ANTLR error listener that prints syntax errors to stderr.

    If a line_map is provided (from the preprocessor), it will map the
    original file and line number for better diagnostics. This keeps
    behavior compatible with previous versions used by the tests.
    """

    def __init__(self, line_map=None):
        super().__init__()
        self.line_map = line_map

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        # column = column + 1
        if self.line_map:
            try:
                file_name, mapped_line = self.line_map.lookup(line)
                print(
                    f"{file_name}::line {mapped_line}:{column} {msg}",
                    file=sys.stderr,
                )
            except Exception:
                print(f"line {line}:{column} {msg}", file=sys.stderr)
        else:
            print(f"line {line}:{column} {msg}", file=sys.stderr)


class ParserWrapper:
    """"""

    def __init__(
        self,
        grammarName,
        startSymbol,
        listener=None,
        debug=False,
        prepro_result=None,
    ):
        self.debug = debug
        self.grammarName = grammarName
        self.startSymbol = startSymbol
        if prepro_result is None:
            filenames, line_map, ifdata_reader = [], None, None
            self.prepro_result = (filenames, line_map, ifdata_reader)
        else:
            filenames, line_map, ifdata_reader = prepro_result
            self.prepro_result = prepro_result
        self.line_map = line_map
        self.lexerModule, self.lexerClass = self._load("Lexer")
        self.parserModule, self.parserClass = self._load("Parser")
        self.listener = listener

    def _load(self, name):
        className = f"{self.grammarName}{name}"
        moduleName = f"pya2l.{className}"
        module = importlib.import_module(moduleName)
        klass = getattr(module, className)
        return (
            module,
            klass,
        )

    def parse(self, input, trace=False):
        self.db = model.A2LDatabase(self.fnbase, debug=self.debug)
        lexer = self.lexerClass(input)
        lexer.removeErrorListeners()
        lexer.addErrorListener(MyErrorListener(self.line_map))
        tokenStream = antlr4.CommonTokenStream(lexer)
        parser = self.parserClass(tokenStream)
        parser.removeErrorListeners()
        parser.addErrorListener(MyErrorListener(self.line_map))
        parser.setTrace(trace)

        parser._interp.debug = True

        meth = getattr(parser, self.startSymbol)
        self._syntaxErrors = parser._syntaxErrors
        tree = meth()
        listener_result = None
        if self.listener:
            self.listener.db = self.db
            listener = self.listener(self.prepro_result)
            walker = antlr4.ParseTreeWalker()
            walker.walk(listener, tree)
            listener_result = listener.result()
        self.db.session.commit()
        return ResultType(self.db, listener_result)

    def parseFromFile(self, filename, encoding="latin-1", trace=False, dbname=":memory:"):
        self.fnbase = dbname
        return self.parse(antlr4.FileStream(filename, encoding), trace)

    def parseFromString(self, buf, encoding="latin-1", trace=False, dbname=":memory:"):
        self.fnbase = dbname
        return self.parse(antlr4.InputStream(buf), trace)

    @staticmethod
    def stringStream(fname, encoding="latin-1"):
        return antlr4.InputStream(open(fname, encoding=encoding).read())

    def _getNumberOfSyntaxErrors(self):
        return self._syntaxErrors

    numberOfSyntaxErrors = property(_getNumberOfSyntaxErrors)


class CustomA2lParser:
    def __init__(self, listener=None, debug=False, prepro_result=None):
        self.debug = debug
        self.startSymbol = "a2lFile"
        if prepro_result is None:
            filenames, line_map, ifdata_reader = [], None, None
            self.prepro_result = (filenames, line_map, ifdata_reader)
        else:
            filenames, line_map, ifdata_reader = prepro_result
            self.prepro_result = prepro_result
        self.line_map = line_map
        self.listener = listener

    def parse(self, filename, encoding="latin-1"):
        self.db = model.A2LDatabase(self.fnbase, debug=self.debug)

        from pya2l.a2lparser import parse

        print("parsing: ", filename)
        tree = parse(filename)

        # parser = self.parserClass(TokenReader(input))
        # parser.removeErrorListeners()
        # parser.addErrorListener(MyErrorListener(self.line_map))
        # meth = getattr(parser, self.startSymbol)
        # tree = meth()

        listener_result = None

        self.listener.db = self.db
        listener = self.listener(self.prepro_result, encoding=encoding)
        walker = antlr4.ParseTreeWalker()
        walker.walk(listener, tree)
        listener_result = listener.result()

        self.db.session.commit()
        return ResultType(self.db, listener_result)

    def parseFromFile(self, filename, encoding="latin-1", dbname=":memory:"):
        self.fnbase = dbname
        return self.parse(filename, encoding=encoding)



#!/usr/bin/env python
"""Tests for the pya2l.api.create module."""

import os
import tempfile
import unittest

import sqlalchemy

import pya2l.model as model
from pya2l.api.create import (
    AxisPtsCreator,
    CharacteristicCreator,
    CompuMethodCreator,
    Creator,
    MeasurementCreator,
)
from pya2l.model import DB_EXTENSION


class TestCreator(unittest.TestCase):
    """Tests for the Creator base class."""

    def setUp(self):
        """Set up test fixture."""
        # Create a temporary database
        fd, self.db_filename = tempfile.mkstemp(suffix=f".{DB_EXTENSION}")
        os.close(fd)

        # Create engine and session
        self.engine = sqlalchemy.create_engine(f"sqlite:///{self.db_filename}")
        model.Base.metadata.create_all(self.engine)
        Session = sqlalchemy.orm.sessionmaker(bind=self.engine)
        self.session = Session()

        # Create a module
        self.module = model.Module(name="TEST_MODULE", longIdentifier="Test Module")
        self.session.add(self.module)
        self.session.commit()

        # Create a creator
        self.creator = Creator(self.session)

    def tearDown(self):
        """Tear down test fixture."""
        self.session.close()
        os.unlink(self.db_filename)

    def test_commit(self):
        """Test commit method."""
        # Add a new object to the session
        obj = model.Unit(name="TEST_UNIT", longIdentifier="Test Unit")
        self.session.add(obj)

        # Commit the session
        self.creator.commit()

        # Check that the object was committed
        self.session.close()
        Session = sqlalchemy.orm.sessionmaker(bind=self.engine)
        session = Session()
        unit = session.query(model.Unit).filter(model.Unit.name == "TEST_UNIT").first()
        self.assertIsNotNone(unit)
        self.assertEqual(unit.longIdentifier, "Test Unit")
        session.close()

    def test_rollback(self):
        """Test rollback method."""
        # Add a new object to the session
        obj = model.Unit(name="TEST_UNIT", longIdentifier="Test Unit")
        self.session.add(obj)

        # Rollback the session
        self.creator.rollback()

        # Check that the object was not committed
        self.session.close()
        Session = sqlalchemy.orm.sessionmaker(bind=self.engine)
        session = Session()
        unit = session.query(model.Unit).filter(model.Unit.name == "TEST_UNIT").first()
        self.assertIsNone(unit)
        session.close()


class TestCompuMethodCreator(unittest.TestCase):
    """Tests for the CompuMethodCreator class."""

    def setUp(self):
        """Set up test fixture."""
        # Create a temporary database
        fd, self.db_filename = tempfile.mkstemp(suffix=f".{DB_EXTENSION}")
        os.close(fd)

        # Create engine and session
        self.engine = sqlalchemy.create_engine(f"sqlite:///{self.db_filename}")
        model.Base.metadata.create_all(self.engine)
        Session = sqlalchemy.orm.sessionmaker(bind=self.engine)
        self.session = Session()

        # Create a module
        self.module = model.Module(name="TEST_MODULE", longIdentifier="Test Module")
        self.session.add(self.module)
        self.session.commit()

        # Create a creator
        self.creator = CompuMethodCreator(self.session)

    def tearDown(self):
        """Tear down test fixture."""
        self.session.close()
        os.unlink(self.db_filename)

    def test_create_compu_method(self):
        """Test create_compu_method method."""
        # Create a computation method
        compu_method = self.creator.create_compu_method(
            name="TEST_CM",
            long_identifier="Test Computation Method",
            conversion_type="LINEAR",
            format_str="%.2f",
            unit="m/s",
            module_name="TEST_MODULE",
        )
        self.creator.commit()

        # Check that the computation method was created
        self.assertIsNotNone(compu_method)
        self.assertEqual(compu_method.name, "TEST_CM")
        self.assertEqual(compu_method.longIdentifier, "Test Computation Method")
        self.assertEqual(compu_method.conversionType, "LINEAR")
        self.assertEqual(compu_method.format, "%.2f")
        self.assertEqual(compu_method.unit, "m/s")
        self.assertEqual(compu_method.module[0], self.module)

    def test_add_coeffs_linear(self):
        """Test add_coeffs_linear method."""
        # Create a computation method
        compu_method = self.creator.create_compu_method(
            name="TEST_CM", long_identifier="Test Computation Method", conversion_type="LINEAR", format_str="%.2f", unit="m/s"
        )

        # Add coefficients
        coeffs_linear = self.creator.add_coeffs_linear(compu_method=compu_method, a=2.0, b=1.0)
        self.creator.commit()

        # Check that the coefficients were added
        self.assertIsNotNone(coeffs_linear)
        self.assertEqual(coeffs_linear.a, 2.0)
        self.assertEqual(coeffs_linear.b, 1.0)
        self.assertEqual(coeffs_linear.compu_method, compu_method)


class TestMeasurementCreator(unittest.TestCase):
    """Tests for the MeasurementCreator class."""

    def setUp(self):
        """Set up test fixture."""
        # Create a temporary database
        fd, self.db_filename = tempfile.mkstemp(suffix=f".{DB_EXTENSION}")
        os.close(fd)

        # Create engine and session
        self.engine = sqlalchemy.create_engine(f"sqlite:///{self.db_filename}")
        model.Base.metadata.create_all(self.engine)
        Session = sqlalchemy.orm.sessionmaker(bind=self.engine)
        self.session = Session()

        # Create a module
        self.module = model.Module(name="TEST_MODULE", longIdentifier="Test Module")
        self.session.add(self.module)
        self.session.commit()

        # Create a creator
        self.creator = MeasurementCreator(self.session)

    def tearDown(self):
        """Tear down test fixture."""
        self.session.close()
        os.unlink(self.db_filename)

    def test_create_measurement(self):
        """Test create_measurement method."""
        # Create a measurement
        measurement = self.creator.create_measurement(
            name="TEST_MEAS",
            long_identifier="Test Measurement",
            datatype="UBYTE",
            conversion="TEST_CM",
            resolution=1,
            accuracy=0.1,
            lower_limit=0.0,
            upper_limit=255.0,
            module_name="TEST_MODULE",
        )
        self.creator.commit()

        # Check that the measurement was created
        self.assertIsNotNone(measurement)
        self.assertEqual(measurement.name, "TEST_MEAS")
        self.assertEqual(measurement.longIdentifier, "Test Measurement")
        self.assertEqual(measurement.datatype, "UBYTE")
        self.assertEqual(measurement.conversion, "TEST_CM")
        self.assertEqual(measurement.resolution, 1)
        self.assertEqual(measurement.accuracy, 0.1)
        self.assertEqual(measurement.lowerLimit, 0.0)
        self.assertEqual(measurement.upperLimit, 255.0)
        self.assertEqual(measurement.module[0], self.module)


class TestAxisPtsCreator(unittest.TestCase):
    """Tests for the AxisPtsCreator class."""

    def setUp(self):
        """Set up test fixture."""
        # Create a temporary database
        fd, self.db_filename = tempfile.mkstemp(suffix=f".{DB_EXTENSION}")
        os.close(fd)

        # Create engine and session
        self.engine = sqlalchemy.create_engine(f"sqlite:///{self.db_filename}")
        model.Base.metadata.create_all(self.engine)
        Session = sqlalchemy.orm.sessionmaker(bind=self.engine)
        self.session = Session()

        # Create a module
        self.module = model.Module(name="TEST_MODULE", longIdentifier="Test Module")
        self.session.add(self.module)
        self.session.commit()

        # Create a creator
        self.creator = AxisPtsCreator(self.session)

    def tearDown(self):
        """Tear down test fixture."""
        self.session.close()
        os.unlink(self.db_filename)

    def test_create_axis_pts(self):
        """Test create_axis_pts method."""
        # Create an axis points
        axis_pts = self.creator.create_axis_pts(
            name="TEST_AXIS",
            long_identifier="Test Axis Points",
            address=0x1000,
            input_quantity="INPUT_QUANTITY",
            deposit_attr="DEPOSIT_ATTR",
            max_diff=0.1,
            conversion="TEST_CM",
            max_axis_points=10,
            lower_limit=0.0,
            upper_limit=100.0,
            module_name="TEST_MODULE",
        )
        self.creator.commit()

        # Check that the axis points was created
        self.assertIsNotNone(axis_pts)
        self.assertEqual(axis_pts.name, "TEST_AXIS")
        self.assertEqual(axis_pts.longIdentifier, "Test Axis Points")
        self.assertEqual(axis_pts.address, 0x1000)
        self.assertEqual(axis_pts.inputQuantity, "INPUT_QUANTITY")
        self.assertEqual(axis_pts.depositAttr, "DEPOSIT_ATTR")
        self.assertEqual(axis_pts.maxDiff, 0.1)
        self.assertEqual(axis_pts.conversion, "TEST_CM")
        self.assertEqual(axis_pts.maxAxisPoints, 10)
        self.assertEqual(axis_pts.lowerLimit, 0.0)
        self.assertEqual(axis_pts.upperLimit, 100.0)
        self.assertEqual(axis_pts.module[0], self.module)


class TestCharacteristicCreator(unittest.TestCase):
    """Tests for the CharacteristicCreator class."""

    def setUp(self):
        """Set up test fixture."""
        # Create a temporary database
        fd, self.db_filename = tempfile.mkstemp(suffix=f".{DB_EXTENSION}")
        os.close(fd)

        # Create engine and session
        self.engine = sqlalchemy.create_engine(f"sqlite:///{self.db_filename}")
        model.Base.metadata.create_all(self.engine)
        Session = sqlalchemy.orm.sessionmaker(bind=self.engine)
        self.session = Session()

        # Create a module
        self.module = model.Module(name="TEST_MODULE", longIdentifier="Test Module")
        self.session.add(self.module)
        self.session.commit()

        # Create a creator
        self.creator = CharacteristicCreator(self.session)

    def tearDown(self):
        """Tear down test fixture."""
        self.session.close()
        os.unlink(self.db_filename)

    def test_create_characteristic(self):
        """Test create_characteristic method."""
        # Create a characteristic
        characteristic = self.creator.create_characteristic(
            name="TEST_CHAR",
            long_identifier="Test Characteristic",
            char_type="VALUE",
            address=0x2000,
            deposit="DEPOSIT",
            max_diff=0.1,
            conversion="TEST_CM",
            lower_limit=0.0,
            upper_limit=100.0,
            module_name="TEST_MODULE",
        )
        self.creator.commit()

        # Check that the characteristic was created
        self.assertIsNotNone(characteristic)
        self.assertEqual(characteristic.name, "TEST_CHAR")
        self.assertEqual(characteristic.longIdentifier, "Test Characteristic")
        self.assertEqual(characteristic.type, "VALUE")
        self.assertEqual(characteristic.address, 0x2000)
        self.assertEqual(characteristic.deposit, "DEPOSIT")
        self.assertEqual(characteristic.maxDiff, 0.1)
        self.assertEqual(characteristic.conversion, "TEST_CM")
        self.assertEqual(characteristic.lowerLimit, 0.0)
        self.assertEqual(characteristic.upperLimit, 100.0)
        self.assertEqual(characteristic.module[0], self.module)

    def test_add_axis_descr(self):
        """Test add_axis_descr method."""
        # Create a characteristic
        characteristic = self.creator.create_characteristic(
            name="TEST_CHAR",
            long_identifier="Test Characteristic",
            char_type="CURVE",
            address=0x2000,
            deposit="DEPOSIT",
            max_diff=0.1,
            conversion="TEST_CM",
            lower_limit=0.0,
            upper_limit=100.0,
        )

        # Add axis description
        axis_descr = self.creator.add_axis_descr(
            characteristic=characteristic,
            attribute="STD_AXIS",
            input_quantity="INPUT_QUANTITY",
            conversion="TEST_CM",
            max_axis_points=10,
            lower_limit=0.0,
            upper_limit=100.0,
        )
        self.creator.commit()

        # Check that the axis description was added
        self.assertIsNotNone(axis_descr)
        self.assertEqual(axis_descr.attribute, "STD_AXIS")
        self.assertEqual(axis_descr.inputQuantity, "INPUT_QUANTITY")
        self.assertEqual(axis_descr.conversion, "TEST_CM")
        self.assertEqual(axis_descr.maxAxisPoints, 10)
        self.assertEqual(axis_descr.lowerLimit, 0.0)
        self.assertEqual(axis_descr.upperLimit, 100.0)
        self.assertEqual(axis_descr.characteristic, characteristic)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: latin-1 -*-
import pytest

from pya2l.aml.classes import AMLPredefinedTypes
from pya2l.aml.classes import map_predefined_type
from pya2l.aml.listener import AMLListener
from pya2l.parserlib import ParserWrapper


def test_type_mapping():
    assert map_predefined_type("char") == AMLPredefinedTypes.PDT_CHAR
    assert map_predefined_type("int") == AMLPredefinedTypes.PDT_INT
    assert map_predefined_type("long") == AMLPredefinedTypes.PDT_LONG
    assert map_predefined_type("uchar") == AMLPredefinedTypes.PDT_UCHAR
    assert map_predefined_type("uint") == AMLPredefinedTypes.PDT_UINT
    assert map_predefined_type("ulong") == AMLPredefinedTypes.PDT_ULONG
    assert map_predefined_type("double") == AMLPredefinedTypes.PDT_DOUBLE
    assert map_predefined_type("float") == AMLPredefinedTypes.PDT_FLOAT


def test_enum_without_tag():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    enum {
      "ADDRESS_GRANULARITY_BYTE" = 1,
      "ADDRESS_GRANULARITY_WORD" = 2,
      "ADDRESS_GRANULARITY_DWORD" = 4
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    enum = res.listener_result.declarations[0].type_definition.type_name.type_
    assert enum.name is None
    enumerators = sorted(enum.enumerators.items(), key = lambda t: t[1])
    assert len(enumerators) == 3
    en0_t, en0_c = enumerators[0]
    en1_t, en1_c = enumerators[1]
    en2_t, en2_c = enumerators[2]
    assert en0_t == "ADDRESS_GRANULARITY_BYTE"
    assert en0_c == 1
    assert en1_t == "ADDRESS_GRANULARITY_WORD"
    assert en1_c == 2
    assert en2_t == "ADDRESS_GRANULARITY_DWORD"
    assert en2_c == 4


def test_enum_with_tag():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    enum checksum {
          "XCP_ADD_11" = 1,
          "XCP_ADD_12" = 2,
          "XCP_ADD_14" = 3,
          "XCP_ADD_22" = 4,
          "XCP_ADD_24" = 5,
          "XCP_ADD_44" = 6,
          "XCP_CRC_16" = 7,
          "XCP_CRC_16_CITT" = 8,
          "XCP_CRC_32" = 9,
          "XCP_USER_DEFINED" = 255
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    enum = res.listener_result.declarations[0].type_definition.type_name.type_
    assert enum.name == "checksum"
    enumerators = sorted(enum.enumerators.items(), key = lambda t: t[1])
    assert len(enumerators) == 10
    en0_t, en0_c = enumerators[0]
    en1_t, en1_c = enumerators[1]
    en2_t, en2_c = enumerators[2]
    en3_t, en3_c = enumerators[3]
    en4_t, en4_c = enumerators[4]
    en5_t, en5_c = enumerators[5]
    en6_t, en6_c = enumerators[6]
    en7_t, en7_c = enumerators[7]
    en8_t, en8_c = enumerators[8]
    en9_t, en9_c = enumerators[9]
    assert en0_t == "XCP_ADD_11"
    assert en0_c == 1
    assert en1_t == "XCP_ADD_12"
    assert en1_c == 2
    assert en2_t == "XCP_ADD_14"
    assert en2_c == 3
    assert en3_t == "XCP_ADD_22"
    assert en3_c == 4
    assert en4_t == "XCP_ADD_24"
    assert en4_c == 5
    assert en5_t == "XCP_ADD_44"
    assert en5_c == 6
    assert en6_t == "XCP_CRC_16"
    assert en6_c == 7
    assert en7_t == "XCP_CRC_16_CITT"
    assert en7_c == 8
    assert en8_t == "XCP_CRC_32"
    assert en8_c == 9
    assert en9_t == "XCP_USER_DEFINED"
    assert en9_c == 255


def test_enum_without_constants():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    enum {
        "PARITY_NONE",
        "PARITY_ODD",
        "PARITY_EVEN"
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    enum = res.listener_result.declarations[0].type_definition.type_name.type_
    assert enum.name is None
    enumerators = sorted(enum.enumerators.items(), key = lambda t: t[1])
    assert len(enumerators) == 3
    en0_t, en0_c = enumerators[0]
    en1_t, en1_c = enumerators[1]
    en2_t, en2_c = enumerators[2]
    assert en0_t == "PARITY_NONE"
    assert en0_c == 0
    assert en1_t == "PARITY_ODD"
    assert en1_c == 1
    assert en2_t == "PARITY_EVEN"
    assert en2_c == 2


def test_enum_one_constant():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    enum {
        "NO_CHECKSUM" = 10,
        "CHECKSUM_BYTE",
        "CHECKSUM_WORD"
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    enum = res.listener_result.declarations[0].type_definition.type_name.type_
    enumerators = sorted(enum.enumerators.items(), key = lambda t: t[1])
    assert enum.name is None
    assert len(enumerators) == 3
    en0_t, en0_c = enumerators[0]
    en1_t, en1_c = enumerators[1]
    en2_t, en2_c = enumerators[2]
    assert en0_t == "NO_CHECKSUM"
    assert en0_c == 10
    assert en1_t == "CHECKSUM_BYTE"
    assert en1_c == 11
    assert en2_t == "CHECKSUM_WORD"
    assert en2_c == 12


def test_enum_unsteady_constants():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    enum {
      "UNIT_1NS" = 0,
      "UNIT_10NS" = 1,
      "UNIT_100NS",
      "UNIT_1US" = 30,
      "UNIT_10US" = 4,
      "UNIT_100US" = 5,
      "UNIT_1MS" = 6,
      "UNIT_10MS" = 7,
      "UNIT_100MS" = 8,
      "UNIT_1S" = 9
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    enum = res.listener_result.declarations[0].type_definition.type_name.type_
    assert enum.name is None
    enumerators = sorted(enum.enumerators.items(), key = lambda t: t[1])
    assert len(enumerators) == 10
    en0_t, en0_c = enumerators[0]
    en1_t, en1_c = enumerators[1]
    en2_t, en2_c = enumerators[2]
    en3_t, en3_c = enumerators[3]
    en4_t, en4_c = enumerators[4]
    en5_t, en5_c = enumerators[5]
    en6_t, en6_c = enumerators[6]
    en7_t, en7_c = enumerators[7]
    en8_t, en8_c = enumerators[8]
    en9_t, en9_c = enumerators[9]
    assert en0_t == "UNIT_1NS"
    assert en0_c == 0
    assert en1_t == "UNIT_10NS"
    assert en1_c == 1
    assert en2_t == "UNIT_100NS"
    assert en2_c == 2
    assert en3_t == "UNIT_10US"
    assert en3_c == 4
    assert en4_t == "UNIT_100US"
    assert en4_c == 5
    assert en5_t == "UNIT_1MS"
    assert en5_c == 6
    assert en6_t == "UNIT_10MS"
    assert en6_c == 7
    assert en7_t == "UNIT_100MS"
    assert en7_c == 8
    assert en8_t == "UNIT_1S"
    assert en8_c == 9
    assert en9_t == "UNIT_1US"
    assert en9_c == 30


def struct(structs):
    assert structs.name is None
    members = structs.members
    assert len(members) == 8
    m0 = members[0]
    tn = m0.value.type_name
    assert tn.type_.type_ == AMLPredefinedTypes.PDT_CHAR
    assert tn.tag is None
    assert tn.name is None
    assert m0.value.array_specifier == [101]
    assert m0.multiple is False
    m1 = members[1]
    tn = m1.value.type_name
    assert tn.type_.type_ == AMLPredefinedTypes.PDT_CHAR
    assert tn.tag is None
    assert tn.name is None
    assert m1.value.array_specifier == [9]
    assert m1.multiple is False
    m2 = members[2]
    tn = m2.value.type_name
    assert tn.type_.type_ == AMLPredefinedTypes.PDT_UINT
    assert tn.tag is None
    assert tn.name is None
    assert m2.value.array_specifier == []
    assert m2.multiple is False
    m3 = members[3]
    tn = m3.value.type_name
    assert tn.tag is None
    assert tn.name is None
    enum = tn.type_
    assert enum.name is None
    enumerators = sorted(enum.enumerators.items(), key = lambda t: t[1])
    assert len(enumerators) == 3
    en0_t, en0_c = enumerators[0]
    assert en0_t == "DAQ"
    assert en0_c == 1
    en1_t, en1_c = enumerators[1]
    assert en1_t == "STIM"
    assert en1_c == 2
    en2_t, en2_c = enumerators[2]
    assert en2_t == "DAQ_STIM"
    assert en2_c == 3
    m4 = members[4]
    assert m4.multiple is False
    tn = m4.value.type_name
    assert tn.tag is None
    assert tn.name is None
    pdt = tn.type_
    assert pdt.type_ == AMLPredefinedTypes.PDT_UCHAR
    m5 = members[5]
    assert m5.multiple is False
    tn = m5.value.type_name
    assert tn.tag is None
    assert tn.name is None
    pdt = tn.type_
    assert pdt.type_ == AMLPredefinedTypes.PDT_UCHAR
    m6 = members[6]
    assert m6.multiple is False
    tn = m6.value.type_name
    assert tn.tag is None
    assert tn.name is None
    pdt = tn.type_
    assert pdt.type_ == AMLPredefinedTypes.PDT_UCHAR
    m7 = members[7]
    assert m7.multiple is False
    tn = m7.value.type_name
    assert tn.tag is None
    assert tn.name is None
    pdt = tn.type_
    assert pdt.type_ == AMLPredefinedTypes.PDT_UCHAR


def test_complex_struct():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    struct {
        char[101];  /* EVENT_CHANNEL_NAME       */
        char[9];    /* EVENT_CHANNEL_SHORT_NAME */
        uint;       /* EVENT_CHANNEL_NUMBER     */
        enum {
          "DAQ" = 1,
          "STIM" = 2,
          "DAQ_STIM" = 3
        };
        uchar;  /* MAX_DAQ_LIST */
        uchar;  /* TIME_CYCLE   */
        uchar;  /* TIME_UNIT    */
        uchar;  /* PRIORITY     */
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    structs = res.listener_result.declarations[0].type_definition.type_name.type_
    struct(structs)


def test_basic_tagged_struct():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    taggedstruct test {
        "SLAVE" ;
        "MASTER" struct {
            uchar;  /* MAX_BS_PGM */
            uchar;  /* MIN_ST_PGM */
        };
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    tagged_struct = struct = res.listener_result.declarations[0].type_definition.type_name.type_
    assert tagged_struct.name == "test"
    members = tagged_struct.members
    assert len(members) == 2
    m0 = members[0]
    assert m0.block_definition is None
    assert m0.multiple is False
    tsd = m0.taggedstruct_definition
    assert tsd.tag == "SLAVE"
    assert tsd.multiple is False
    assert tsd.member is None
    m1 = members[1]
    assert m1.block_definition is None
    tsd = m1.taggedstruct_definition
    assert tsd.tag == "MASTER"
    assert tsd.multiple is False
    member = tsd.member
    tn = member.type_name
    assert tn.tag is None
    assert tn.name is None
    struct = tn.type_
    assert struct.name is None
    members = struct.members
    assert len(members) == 2
    m0 = members[0]
    assert m0.multiple is False
    tn = m0.value.type_name
    assert tn.tag is None
    assert tn.name is None
    assert tn.type_.type_ == AMLPredefinedTypes.PDT_UCHAR
    m1 = members[1]
    assert m1.multiple is False
    tn = m1.value.type_name
    assert tn.tag is None
    assert tn.name is None
    assert tn.type_.type_ == AMLPredefinedTypes.PDT_UCHAR


def test_basic_tagged_union():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    taggedunion Daq_Event {
        "FIXED_EVENT_LIST" taggedstruct {
            ("EVENT" uint)*;
        };
        "VARIABLE" taggedstruct {
            block "AVAILABLE_EVENT_LIST" taggedstruct {
                ("EVENT" uint)*;
            };
            block "DEFAULT_EVENT_LIST" taggedstruct {
                ("EVENT" uint)*;
            };
        };
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    tagged_union = res.listener_result.declarations[0].type_definition.type_name.type_
    assert tagged_union.name == "Daq_Event"
    members = tagged_union.members
    assert len(members) == 2
    m0 = members[0]
    assert m0.tag == "FIXED_EVENT_LIST"
    assert m0.block_definition is None
    mem = m0.member
    tn = mem.type_name.type_
    assert tn.name is None
    assert len(tn.members) == 1
    mem = tn.members[0]
    assert mem.multiple
    assert mem.block_definition is None
    td = mem.taggedstruct_definition
    assert td.tag == "EVENT"
    assert td.multiple is False
    tn = td.member.type_name
    assert tn.tag is None
    assert tn.name is None
    assert tn.type_.type_ == AMLPredefinedTypes.PDT_UINT
    m1 = members[1]
    assert m1.tag == "VARIABLE"
    member = m1.member
    tn = member.type_name
    assert tn.tag is None
    assert tn.name is None
    tp = tn.type_
    assert tp.name is None
    members = tp.members
    assert len(members) == 2
    m0 = members[0]
    assert m0.multiple is False
    assert m0.taggedstruct_definition is None
    bd = m0.block_definition
    assert bd.tag == "AVAILABLE_EVENT_LIST"
    assert bd.member is None
    tn = bd.type_name
    assert tn.tag is None
    assert tn.name is None
    tst = tn.type_
    assert tst.name is None
    ms = tst.members
    assert len(ms) == 1
    m = ms[0]
    assert m.block_definition is None
    assert m.multiple
    tsd = m.taggedstruct_definition
    assert tsd.multiple is False
    assert tsd.tag == "EVENT"
    tn = tsd.member.type_name
    assert tn.tag is None
    assert tn.name is None
    assert tn.type_.type_ == AMLPredefinedTypes.PDT_UINT
    m1 = members[1]
    assert m1.multiple is False
    assert m1.taggedstruct_definition is None
    bd = m1.block_definition
    assert bd.tag == "DEFAULT_EVENT_LIST"
    assert bd.member is None
    tn = bd.type_name
    assert tn.tag is None
    assert tn.name is None
    tst = tn.type_
    assert tst.name is None
    ms = tst.members
    assert len(ms) == 1
    m = ms[0]
    assert m.block_definition is None
    assert m.multiple
    tsd = m.taggedstruct_definition
    assert tsd.multiple is False
    assert tsd.tag == "EVENT"
    tn = tsd.member.type_name
    assert tn.tag is None
    assert tn.name is None
    assert tn.type_.type_ == AMLPredefinedTypes.PDT_UINT


def block_def(block_definitions):
    bd = block_definitions
    assert bd.tag == "STIM"
    assert bd.member is None
    tn = bd.type_name
    assert tn.tag is None
    assert tn.name is None
    tp = bd.type_name
    assert tp.tag is None
    assert tp.name is None
    st = tp.type_
    assert st.name is None
    members = st.members
    assert len(members) == 3
    m0 = members[0].value
    tn = m0.type_name
    assert tn.tag is None
    assert tn.name is None
    enum = tn.type_
    assert enum.name is None
    enumerators = sorted(enum.enumerators.items(), key = lambda t: t[1])
    assert len(enumerators) == 4
    en0_t, en0_c = enumerators[0]
    assert en0_t == "GRANULARITY_ODT_ENTRY_SIZE_STIM_BYTE"
    assert en0_c == 1
    en1_t, en1_c  = enumerators[1]
    assert en1_t == "GRANULARITY_ODT_ENTRY_SIZE_STIM_WORD"
    assert en1_c == 2
    en2_t, en2_c = enumerators[2]
    assert en2_t == "GRANULARITY_ODT_ENTRY_SIZE_STIM_DWORD"
    assert en2_c == 4
    en3_t, en3_c = enumerators[3]
    assert en3_t == "GRANULARITY_ODT_ENTRY_SIZE_STIM_DLONG"
    assert en3_c == 8
    m1 = members[1].value
    tn = m1.type_name
    assert tn.tag is None
    assert tn.name is None
    assert tn.type_.type_ == AMLPredefinedTypes.PDT_UCHAR
    m2 = members[2].value
    tn = m2.type_name
    assert tn.tag is None
    assert tn.name is None
    tp = tn.type_
    assert tp.name is None
    members = tp.members
    assert len(members) == 1
    mem = members[0]
    assert mem.block_definition is None
    assert mem.multiple is False
    tsd = mem.taggedstruct_definition
    assert tsd.tag == "BIT_STIM_SUPPORTED"
    assert tsd.member is None
    assert tsd.multiple is False


def test_basic_block_definition():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    block "STIM" struct {
        enum {
          "GRANULARITY_ODT_ENTRY_SIZE_STIM_BYTE" = 1,
          "GRANULARITY_ODT_ENTRY_SIZE_STIM_WORD" = 2,
          "GRANULARITY_ODT_ENTRY_SIZE_STIM_DWORD" = 4,
          "GRANULARITY_ODT_ENTRY_SIZE_STIM_DLONG" = 8
        };
        uchar;  /* MAX_ODT_ENTRY_SIZE_STIM */
        taggedstruct {
          "BIT_STIM_SUPPORTED" ;
        };
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    block_definition = res.listener_result.declarations[0].block_definition
    block_def(block_definition)


def test_struct_referrers():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    struct FLX_Parameters {
        taggedunion {
            block "INITIAL_CMD_BUFFER" struct buffer;
        };
        taggedunion {
            block "INITIAL_RES_ERR_BUFFER" struct buffer;
        };
        taggedstruct {
            (block "POOL_BUFFER" struct buffer)*;
        };
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    struct_t = res.listener_result.declarations[0].type_definition.type_name.type_
    assert struct_t.name == "FLX_Parameters"
    members = struct_t.members
    assert len(members) == 3
    m0 = members[0]
    assert m0.multiple is False
    tn = m0.value.type_name
    assert tn.tag is None
    assert tn.name is None
    tp = tn.type_
    assert tp.name is None
    ms = tp.members
    assert len(ms) == 1
    m = ms[0]
    assert m.tag is None
    assert m.member is None
    bd = m.block_definition
    assert bd.tag == "INITIAL_CMD_BUFFER"
    tn = bd.type_name
    assert tn.tag is None
    assert tn.name is None
    assert tn.type_.identifier == "buffer"
    assert tn.type_.category == "StructType"
    assert bd.member is None
    assert bd.multiple == False
    m1 = members[1]
    assert m1.multiple is False
    tn = m1.value.type_name
    assert tn.tag is None
    assert tn.name is None
    tp = tn.type_
    assert tp.name is None
    ms = tp.members
    assert len(ms) == 1
    m = ms[0]
    assert m.tag is None
    assert m.member is None
    bd = m.block_definition
    assert bd.tag == "INITIAL_RES_ERR_BUFFER"
    tn = bd.type_name
    assert tn.tag is None
    assert tn.type_.identifier == "buffer"
    assert tn.type_.category == "StructType"
    assert bd.member is None
    assert bd.multiple == False
    m2 = members[2]
    assert m2.multiple is False
    tn = m2.value.type_name
    assert tn.tag is None
    assert tn.name is None
    tst = tn.type_
    assert tst.name is None
    members = tst.members
    assert len(members) == 1
    mem = members[0]
    assert mem.multiple
    assert mem.taggedstruct_definition is None
    bd = mem.block_definition
    assert bd.tag == "POOL_BUFFER"
    assert bd.member is None
    tn = bd.type_name
    assert tn.tag is None
    assert tn.type_.identifier == "buffer"
    assert tn.type_.category == "StructType"
    assert bd.member is None
    assert bd.multiple == False

def test_taggedstruct_referrers():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
        struct {
            taggedstruct Common_Parameters;  /* default parameters */
            taggedstruct bus_systems {
                block "XCP_ON_CAN" struct {
                    struct CAN_Parameters;  /* specific for CAN */
                    taggedstruct Common_Parameters;  /* overruling of default */
                };
                block "XCP_ON_SxI" struct {
                    struct SxI_Parameters;  /* specific for SxI */
                    taggedstruct Common_Parameters;  /* overruling of default */
                };
                block "XCP_ON_TCP_IP" struct {
                    struct TCP_IP_Parameters;  /* specific for TCP_IP */
                    taggedstruct Common_Parameters;  /* overruling of default */
                };
                block "XCP_ON_UDP_IP" struct {
                    struct UDP_Parameters;  /* specific for UDP */
                    taggedstruct Common_Parameters;  /* overruling of default */
                };
                block "XCP_ON_USB" struct {
                    struct USB_Parameters;  /* specific for USB      */
                    taggedstruct Common_Parameters;  /* overruling of default */
                };
                block "XCP_ON_FLX" struct {
                    struct FLX_Parameters;  /* specific for FlexRay  */
                    taggedstruct Common_Parameters;  /* overruling of default */
                };
            };
        };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    tagged_struct = res.listener_result.declarations[0].type_definition.type_name.type_
    members = tagged_struct.members
    mem = members[0]
    assert mem.multiple == False
    mem = mem.value.type_name
    assert mem.tag is None
    assert mem.name is None
    assert mem.type_.category == "TaggedStructType"
    assert mem.type_.identifier == "Common_Parameters"

    mem = members[1]
    assert mem.multiple == False
    mem = mem.value.type_name

    assert mem.name == "bus_systems"
    assert mem.type_.name == "bus_systems"
    members = mem.type_.members
    assert len(members) == 6

    m0 = members[0]
    assert m0.multiple is False
    assert m0.taggedstruct_definition is None
    bd = m0.block_definition
    assert bd.tag == "XCP_ON_CAN"
    assert bd.member is None
    tn = bd.type_name
    assert tn.tag is None
    assert tn.name is None
    tp = tn.type_
    assert tp.name is None
    ms = members

    assert len(ms) == 6
    m00 = ms[0]


def test_taggedunion_referrers():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """
    /begin A2ML
    taggedstruct Common_Parameters {
        block "PROTOCOL_LAYER" struct Protocol_Layer;
        block "SEGMENT" struct Segment;
        block "DAQ" struct Daq;
        block "PAG" struct Pag;
        block "PGM" struct Pgm;
        block "DAQ_EVENT" taggedunion Daq_Event;
    };
    /end A2ML
    """
    res = parser.parseFromString(DATA)
    tagged_struct = res.listener_result.declarations[0].type_definition.type_name.type_
    ts = tagged_struct
    assert ts.name == "Common_Parameters"
    members = ts.members
    assert len(members) == 6
    m0 = members[0]
    assert m0.multiple is False
    assert m0.taggedstruct_definition is None

    bd = m0.block_definition
    assert bd.member is None
    assert bd.tag == "PROTOCOL_LAYER"
    tn = bd.type_name
    assert tn.tag is None
    assert tn.type_.identifier == "Protocol_Layer"
    m1 = members[1]
    assert m1.multiple is False
    assert m1.taggedstruct_definition is None

    bd = m1.block_definition
    assert bd.member is None
    assert bd.tag == "SEGMENT"
    tn = bd.type_name
    assert tn.type_.identifier == "Segment"
    m2 = members[2]
    assert m2.multiple is False
    assert m2.taggedstruct_definition is None

    bd = m2.block_definition
    assert bd.member is None
    assert bd.tag == "DAQ"
    tn = bd.type_name
    assert tn.type_.identifier == "Daq"
    m3 = members[3]
    assert m3.multiple is False
    assert m3.taggedstruct_definition is None

    bd = m3.block_definition
    assert bd.member is None
    assert bd.tag == "PAG"
    tn = bd.type_name
    assert tn.type_.identifier == "Pag"
    m4 = members[4]
    assert m4.multiple is False
    assert m4.taggedstruct_definition is None

    bd = m4.block_definition
    assert bd.member is None
    assert bd.tag == "PGM"
    tn = bd.type_name
    assert tn.tag is None
    assert tn.type_.identifier == "Pgm"
    m5 = members[5]
    assert m5.multiple is False
    assert m5.taggedstruct_definition is None

    bd = m5.block_definition
    assert bd.member is None
    assert bd.tag == "DAQ_EVENT"
    tn = bd.type_name
    assert tn.tag is None
    assert tn.type_.identifier == "Daq_Event"

def test_aml_13():
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    DATA = """/begin A2ML
block "IF_DATA" taggedunion if_data {
"ASAP1B_Bypass" taggedstruct {
(block "SOURCE" struct {
struct {
char[101];
int;
long;
};
taggedstruct {
block "QP_BLOB" struct {
uint;
};
};
})*;
block "TP_BLOB" struct {
uint;
uint;
};
block "DP_BLOB" struct {
};
block "PA_BLOB" struct {
};
block "KP_BLOB" struct {
taggedstruct {
"BUFFER_OFFSET"    char[256];
"SOURCE_ID"        char[256];
"BIT_OFFSET"       char[256];
"POSSIBLE_SOURCES" ( uint )*;
};
};
};
"ASAP1B_CCP" taggedstruct {
(block "SOURCE" struct {
struct {
char [101];
int;
long;
};
taggedstruct {
"DISPLAY_IDENTIFIER" char[32];
block "QP_BLOB" struct {
uint;
taggedstruct {
"LENGTH" uint;
"CAN_ID_VARIABLE";
"CAN_ID_FIXED" ulong;
("RASTER" uchar; )*;
("EXCLUSIVE" int; )*;
"REDUCTION_ALLOWED";
"FIRST_PID" uchar;
};
};
};
}; )*;
(block "RASTER" struct {
char [101];
char [9];
uchar;
int;
long;
taggedstruct {
("EXCLUSIVE" uchar; )*;
};
}; )*;
(block "EVENT_GROUP" struct {
char [101];
char [9];
taggedstruct {
("RASTER" uchar; )*;
};
}; )*;
block "SEED_KEY" struct {
char[256];
char[256];
char[256];
};
block "CHECKSUM" struct {
char[256];
};
block "TP_BLOB" struct {
uint;
uint;
ulong;
ulong;
uint;
uint;
taggedstruct {
block "CAN_PARAM" struct {
uint;
uchar;
uchar;
};
"BAUDRATE" ulong;
"SAMPLE_POINT" uchar;
"SAMPLE_RATE" uchar;
"BTL_CYCLES" uchar;
"SJW" uchar;
"SYNC_EDGE" enum {
"SINGLE" = 0,
"DUAL" = 1
};
"DAQ_MODE" enum {
"ALTERNATING" = 0,
"BURST" = 1
};
"BYTES_ONLY";
"RESUME_SUPPORTED";
"STORE_SUPPORTED";
"CONSISTENCY" enum {
"DAQ" = 0,
"ODT" = 1
};
"ADDRESS_EXTENSION" enum {
"DAQ" = 0,
"ODT" = 1
};
block "CHECKSUM_PARAM" struct {
uint;
ulong;
taggedstruct {
"CHECKSUM_CALCULATION" enum {
"ACTIVE_PAGE" = 0,
"BIT_OR_WITH_OPT_PAGE" = 1
};
};
};
(block "DEFINED_PAGES" struct {
struct {
uint;
char[101];
uint;
ulong;
ulong;
};
taggedstruct {
"RAM";
"ROM";
"FLASH";
"EEPROM";
"RAM_INIT_BY_ECU";
"RAM_INIT_BY_TOOL";
"AUTO_FLASH_BACK";
"FLASH_BACK";
"DEFAULT";
};
}; ) *;
( "OPTIONAL_CMD"  uint; )*;
};
};
"DP_BLOB" struct {
uint;
ulong;
ulong;
};
"KP_BLOB" struct {
uint;
ulong;
ulong;
taggedstruct {
("RASTER" uchar; )*;
};
};
};
"ETK" taggedstruct
{
("ADDRESS_MAPPING" struct
{
ulong;
ulong;
ulong;
}
)*;
(block "SOURCE" struct
{
struct
{
char [100];
int;
long;
};
taggedstruct
{
"QP_BLOB" struct
{
uint;
uint;
ulong;
ulong;
ulong;
uint;
uint;
ulong;
};
};
})*;
block "TP_BLOB" struct
{
ulong;
ulong;
ulong;
block "DISTAB_CFG" struct {
uint;
uint;
uint;
ulong;
ulong;
taggedstruct {
"TRG_MOD" ( uchar; )*;
};
};
taggedstruct {
"CODE_CHK" struct {
ulong;
uint;
ulong;
uint;
};
"ETK_CFG" ( uchar; )*;
( "EMU_DATA" struct {
ulong;
ulong;
enum {
"INTERN"  = 0,
"EXTERN"  = 1
};
ulong;
ulong;
ulong;
ulong;
ulong;
})*;
( "EMU_CODE" struct {
ulong;
ulong;
enum {
"INTERN"  = 0,
"EXTERN"  = 1
};
ulong;
ulong;
ulong;
ulong;
ulong;
})*;
( "EMU_RAM" struct {
ulong;
ulong;
enum {
"INTERN"  = 0,
"EXTERN"  = 1
};
ulong;
ulong;
ulong;
ulong;
ulong;
})*;
( "RESERVED" struct {
ulong;
ulong;
enum {
"INTERN"  = 0,
"EXTERN"  = 1
};
ulong;
ulong;
ulong;
ulong;
ulong;
})*;
"ETK_MAILBOX" struct {
ulong;
enum {
"CODE"  = 1,
"DATA"  = 2,
"EXRAM" = 3
};
ulong;
ulong;
enum {
"CODE"  = 1,
"DATA"  = 2,
"EXRAM" = 3
};
ulong;
ulong;
};
}
};
"DP_BLOB" struct
{
ulong;
ulong;
};
"KP_BLOB" struct
{
ulong;
uint;
uint;
};
};
"ASAP1B_KWP2000"  taggedstruct {
("ADDRESS_MAPPING" struct
{
ulong;
ulong;
ulong;
}
)*;
( "SEED_KEY"
char[256]
)*;
(block "SOURCE" struct {
struct {
char [100];
int;
long;
};
taggedstruct {
"QP_BLOB" struct {
uint;
enum {
"ADDRESSMODE"  = 1,
"BLOCKMODE"    = 2
};
uint;
uint;
uint;
};
};
}
)*;
block "TP_BLOB" struct {
uint;
uint;
uint;
enum {
"WuP"   = 1,
"5Baud" = 2
};
enum {
"MSB_FIRST" = 1,
"MSB_LAST"  = 2
};
uint;
taggedstruct {
"DATA_ACCESS" struct {
ulong;
ulong;
uint;
uint;
uint;
uint;
};
block "CHECKSUM" struct {
ulong;
uint;
uint;
enum {
"RequestRoutineResults" = 0,
"StartRoutine"          = 1,
"CodedResult"			  = 2
};
taggedstruct {
"RNC_RESULT" ( uchar )*;
};
};
block "FLASH_COPY" struct {
enum {
"NOFLASHBACK"   = 0x00,
"AUTOFLASHBACK" = 0x40,
"TOOLFLASHBACK" = 0x80
};
uint;
enum {
"RequestRoutineResults" = 0,
"StartRoutine"          = 1,
"CodedResult"			  = 2
};
enum {
"RAM_InitByECU"  = 0x10,
"RAM_InitByTool" = 0x20
};
uint;
taggedstruct {
"COPY_FRAME" ( uchar )*;
"RNC_RESULT" ( uchar )*;
"COPY_PARA" ( uchar )*;
};
};
( block "DIAG_BAUD" struct {
ulong;
uint;
taggedstruct {
"BD_PARA" ( uchar )*;
};
}
)*;
( "TIME_DEF" struct {
uint;
uint;
uint;
uint;
uint;
uint;
}
)*;
( "SECURITY_ACCESS" struct {
uint;
uint;
uint;
}
)*;
block "PAGE_SWITCH" struct {
enum {
"ESCAPE_CODE"   = 0x80,
"LOCAL_ROUTINE" = 0x31
};
taggedstruct {
"ESCAPE_CODE_PARA"  ( uchar )*;
block "ROUTINE_PARA" struct {
uint;
enum {
"RequestRoutineResults" = 0,
"StartRoutine"          = 1,
"CodedResult" 	      = 2
};
taggedstruct {
"RNC_RESULT" ( uchar )*;
};
};
"PAGE_CODE" ( uchar )*;
};
};
};
};
"DP_BLOB" struct {
ulong;
ulong;
};
"KP_BLOB" struct {
ulong;
enum {
"INTERN"  = 0,
"EXTERN"  = 1
};
uint;
};
};
"ASAP1B_MCMESS" taggedstruct
{
(block "SOURCE" struct
{
struct
{
char [100];
int;
long;
};
taggedstruct
{
"QP_BLOB" struct
{
enum
{
"HANDSHAKE" = 0x00,
"ANGLE"     = 0x08,
"TIME"      = 0x0C
};
uint;
};
};
}
)*;
block "TP_BLOB" struct
{
uint;
taggedstruct
{
"KOMKEN"      uint;
"MCTGT"       long;
"MCINI"       enum
{
"MASSE"  = 0,
"KW2000" = 1
};
"WFSIDLE0"    uint;
"WFSIDLE1"    uint;
"WFS5B"       uint;
"NOOFCYL"     uint;
"CYLCNAME"    char[20];
"CYLSEQU"     char[30];
"XRAMOF"      uint;
block "CHECKSUM_PARAM" struct
{
ulong;
taggedstruct
{
"CHECKSUM_CALCULATION"
enum
{
"ACTIVE_PAGE",
"WORKING_PAGE"
};
};
};
(block "DEFINED_PAGES" struct
{
uint;
char[101];
taggedunion
{
"RAM" struct
{
enum
{
"RAM_INIT_BY_ECU",
"RAM_INIT_BY_TOOL"
};
taggedstruct
{
"DEFAULT";
};
};
"ROM" taggedstruct
{
"DEFAULT";
};
"FLASH" struct
{
enum
{
"NO_FLASH_BACK",
"AUTO_FLASH_BACK",
"FLASH_BACK"
};
taggedstruct
{
"DEFAULT";
};
};
"EEPROM" struct
{
enum
{
"NO_FLASH_BACK",
"AUTO_FLASH_BACK",
"FLASH_BACK"
};
taggedstruct
{
"DEFAULT";
};
};
};
})*;
"BYTE_ORDER" enum
{
"MSB_FIRST",
"MSB_LAST"
};
"THREE_BYTE_ADDRESSES";
"COPY_FLASH_TO_RAM_BY_COMMAND";
};
};
block "KP_BLOB" struct
{
long;
enum
{
"INTERN",
"EXTERN"
};
uint;
taggedstruct
{
"PSEUDO_ADR" uint;
"VS_DEF"     char[20];
};
};
"DP_BLOB" struct
{
ulong;
ulong;
};
};
};
/end A2ML"""
    res = parser.parseFromString(DATA)


def test_taggedstruct2():
    DATA = """
    /begin A2ML
    taggedstruct {
        ("EXCLUSIVE" uchar; )*;
        "POSSIBLE_SOURCES" ( uint )*;
    };
    /end A2ML
    """
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    res = parser.parseFromString(DATA)


def test_taggedstruct3():
    DATA = """
    /begin A2ML
    taggedstruct {
    (block "RASTER" struct {
    char [101];
    char [9];
    uchar;
    int;
    long;
    taggedstruct {
    ("EXCLUSIVE" uchar; )*;
    };
    }; )*;
    };
    /end A2ML
    """
    parser = ParserWrapper("aml", "amlFile", AMLListener)
    res = parser.parseFromString(DATA)
